@@ -1,7 +1,7 @@
 /* crypto.c: crytography-related functions
-   Copyright (c) 2002-2003 Philip Kendall
+   Copyright (c) 2002-2005 Philip Kendall
 
-   $Id: crypto.c,v 1.8 2003/10/06 12:55:12 pak21 Exp $
+   $Id: crypto.c 2890 2007-05-26 19:31:43Z zubzero $
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -13,14 +13,13 @@
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
 
-   You should have received a copy of the GNU General Public License
-   along with this program; if not, write to the Free Software
-   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
+   You should have received a copy of the GNU General Public License along
+   with this program; if not, write to the Free Software Foundation, Inc.,
+   51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 
    Author contact information:
 
-   E-mail: pak21-fuse@srcf.ucam.org
-   Postal address: 15 Crescent Road, Wokingham, Berks, RG40 2DB, England
+   E-mail: philip-fuse@shadowmagic.org.uk
 
 */
 
@@ -117,7 +116,7 @@ static libspectrum_error
 get_hash( gcry_sexp_t *hash, const libspectrum_byte *data, size_t data_length )
 {
   gcry_error_t error;
-  char *digest; size_t digest_length;
+  unsigned char *digest; size_t digest_length;
   gcry_mpi_t hash_mpi;
   
   digest_length = gcry_md_get_algo_dlen( HASH_ALGORITHM );
@@ -169,15 +168,20 @@ create_key( gcry_sexp_t *s_key, libspectrum_rzx_dsa_key *key,
 
   for( i=0; i<MPI_COUNT; i++ ) mpis[i] = NULL;
 
-  error = gcry_mpi_scan( &mpis[0], GCRYMPI_FMT_HEX, key->p, 0, NULL );
+    error = gcry_mpi_scan( &mpis[0], GCRYMPI_FMT_HEX, (unsigned char*)key->p,
+			   0, NULL );
   if( !error ) 
-    error = gcry_mpi_scan( &mpis[1], GCRYMPI_FMT_HEX, key->q, 0, NULL );
+    error = gcry_mpi_scan( &mpis[1], GCRYMPI_FMT_HEX, (unsigned char*)key->q,
+			   0, NULL );
   if( !error )
-    error = gcry_mpi_scan( &mpis[2], GCRYMPI_FMT_HEX, key->g, 0, NULL );
+    error = gcry_mpi_scan( &mpis[2], GCRYMPI_FMT_HEX, (unsigned char*)key->g,
+			   0, NULL );
   if( !error )
-    error = gcry_mpi_scan( &mpis[3], GCRYMPI_FMT_HEX, key->y, 0, NULL );
+    error = gcry_mpi_scan( &mpis[3], GCRYMPI_FMT_HEX, (unsigned char*)key->y,
+			   0, NULL );
   if( !error && secret_key )
-    error = gcry_mpi_scan( &mpis[4], GCRYMPI_FMT_HEX, key->x, 0, NULL );
+    error = gcry_mpi_scan( &mpis[4], GCRYMPI_FMT_HEX, (unsigned char*)key->x,
+			   0, NULL );
 
   if( error ) {
     libspectrum_print_error( LIBSPECTRUM_ERROR_LOGIC,
@@ -251,7 +255,7 @@ serialise_mpis( libspectrum_byte **signature, size_t *signature_length,
 {
   gcry_error_t error;
   size_t length, length_s;
-  char *ptr;
+  unsigned char *ptr;
 
   error = gcry_mpi_print( GCRYMPI_FMT_PGP, NULL, 0, &length, r );
   if( error ) {
@@ -303,7 +307,7 @@ serialise_mpis( libspectrum_byte **signature, size_t *signature_length,
 #include <stdio.h>
 
 libspectrum_error
-libspectrum_verify_signature( libspectrum_rzx_signature *signature,
+libspectrum_verify_signature( libspectrum_signature *signature,
 			      libspectrum_rzx_dsa_key *key )
 {
   libspectrum_error error;
@@ -351,7 +355,7 @@ libspectrum_verify_signature( libspectrum_rzx_signature *signature,
 }
 
 libspectrum_error
-libspectrum_signature_free( libspectrum_rzx_signature *signature )
+libspectrum_signature_free( libspectrum_signature *signature )
 {
   gcry_mpi_release( signature->r ); gcry_mpi_release( signature->s );
 
