@@ -1,7 +1,7 @@
 /* dck.c: Routines for handling Warajevo DCK files
    Copyright (c) 2003 Darren Salt, Fredrick Meunier
 
-   $Id: dck.c,v 1.3.2.1 2004/01/26 22:41:24 pak21 Exp $
+   $Id: dck.c 2890 2007-05-26 19:31:43Z zubzero $
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -13,19 +13,20 @@
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
 
-   You should have received a copy of the GNU General Public License
-   along with this program; if not, write to the Free Software
-   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
+   You should have received a copy of the GNU General Public License along
+   with this program; if not, write to the Free Software Foundation, Inc.,
+   51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 
    Author contact information:
 
-   E-mail: pak21-fuse@srcf.ucam.org
-     Post: 15 Crescent Road, Wokingham, Berks, RG40 2DB, England
+   E-mail: philip-fuse@shadowmagic.org.uk
 
    Darren: linux@youmustbejoking.demon.co.uk
 
 */
 
+#include <config.h>
+
 #include <stdlib.h>
 #include <string.h>
 
