@@ -1,8 +1,8 @@
 /* internals.h: functions which need to be called inter-file by libspectrum
                 routines, but not by user code
-   Copyright (c) 2001-2002 Philip Kendall, Darren Salt
+   Copyright (c) 2001-2007 Philip Kendall, Darren Salt
 
-   $Id: internals.h,v 1.17 2003/10/20 10:34:40 pak21 Exp $
+   $Id: internals.h 3370 2007-11-30 07:48:32Z zubzero $
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -14,14 +14,13 @@
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
 
-   You should have received a copy of the GNU General Public License
-   along with this program; if not, write to the Free Software
-   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
+   You should have received a copy of the GNU General Public License along
+   with this program; if not, write to the Free Software Foundation, Inc.,
+   51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 
    Author contact information:
 
-   E-mail: pak21-fuse@srcf.ucam.org
-   Postal address: 15 Crescent Road, Wokingham, Berks, RG40 2DB, England
+   E-mail: philip-fuse@shadowmagic.org.uk
 
 */
 
@@ -44,10 +43,19 @@
 #define GCC_PRINTF( fmtstring, args )
 #endif				/* #ifdef __GNUC__ */
 
-/* Win32 systems have _strcmpi, but not strcasecmp */
-#if !defined(HAVE_STRCASECMP) && defined(HAVE__STRCMPI)
-#define strcasecmp _strcmpi
-#endif		/* #if !defined(HAVE_STRCASECMP) && defined(HAVE__STRCMPI) */
+/* On Win32 systems, map snprintf -> _snprintf, strcasecmp -> _stricmp and
+   strncasecmp -> _strnicmp */
+#if !defined(HAVE_SNPRINTF) && defined(HAVE__SNPRINTF)
+#define snprintf _snprintf
+#endif		/* #if !defined(HAVE_SNPRINTF) && defined(HAVE__SNPRINTF) */
+
+#if !defined(HAVE_STRCASECMP) && defined(HAVE__STRICMP)
+#define strcasecmp _stricmp
+#endif		/* #if !defined(HAVE_STRCASECMP) && defined(HAVE__STRICMP) */
+
+#if !defined(HAVE_STRNCASECMP) && defined(HAVE__STRNICMP)
+#define strncasecmp _strnicmp
+#endif		/* #if !defined(HAVE_STRNCASECMP) && defined(HAVE__STRNICMP) */
 
 /* Print using the user-provided error function */
 libspectrum_error
@@ -87,17 +95,36 @@ libspectrum_error
 libspectrum_bzip2_inflate( const libspectrum_byte *bzptr, size_t bzlength,
 			   libspectrum_byte **outptr, size_t *outlength );
 
+/* The TZX file signature */
+
+extern const char *libspectrum_tzx_signature;
+
 /* Convert a 48K memory dump into separate RAM pages */
 
 int libspectrum_split_to_48k_pages( libspectrum_snap *snap,
 				    const libspectrum_byte* data );
 
+/* Sizes of some of the arrays in the snap structure */
+#define SNAPSHOT_RAM_PAGES 16
+#define SNAPSHOT_SLT_PAGES 256
+#define SNAPSHOT_ZXATASP_PAGES 32
+#define SNAPSHOT_ZXCF_PAGES 64
+#define SNAPSHOT_DOCK_EXROM_PAGES 8
+#define SNAPSHOT_JOYSTICKS 7
+
+/* Get memory for a snap */
+
+libspectrum_error libspectrum_snap_alloc_internal( libspectrum_snap **snap );
+
 /* Format specific snapshot routines */
 
 libspectrum_error
 libspectrum_plusd_read( libspectrum_snap *snap,
 			const libspectrum_byte *buffer, size_t buffer_length );
 libspectrum_error
+internal_sna_read( libspectrum_snap *snap,
+		   const libspectrum_byte *buffer, size_t buffer_length );
+libspectrum_error
 libspectrum_sna_write( libspectrum_byte **buffer, size_t *length,
 		       int *out_flags, libspectrum_snap *snap, int in_flags );
 libspectrum_error
@@ -114,12 +141,64 @@ libspectrum_szx_write( libspectrum_byte **buffer, size_t *length,
 		       int *out_flags, libspectrum_snap *snap,
 		       libspectrum_creator *creator, int in_flags );
 libspectrum_error
+internal_z80_read( libspectrum_snap *snap,
+		   const libspectrum_byte *buffer, size_t buffer_length );
+libspectrum_error
 libspectrum_z80_write2( libspectrum_byte **buffer, size_t *length,
 			int *out_flags, libspectrum_snap *snap, int in_flags );
 libspectrum_error
 libspectrum_zxs_read( libspectrum_snap *snap,
 		      const libspectrum_byte *buffer, size_t buffer_length );
 
+/* Tape routines */
+
+void libspectrum_tape_block_zero( libspectrum_tape_block *block );
+
+libspectrum_error libspectrum_tape_block_read_symbol_table_parameters(
+  libspectrum_tape_block *block, int pilot, const libspectrum_byte **ptr );
+
+libspectrum_error
+libspectrum_tape_block_read_symbol_table(
+  libspectrum_tape_generalised_data_symbol_table *table,
+  const libspectrum_byte **ptr, size_t length );
+
+/* Format specific tape routines */
+  
+libspectrum_error
+internal_tap_read( libspectrum_tape *tape, const libspectrum_byte *buffer,
+		   const size_t length, libspectrum_id_t type );
+
+libspectrum_error
+internal_tap_write( libspectrum_byte **buffer, size_t *length,
+		    libspectrum_tape *tape, libspectrum_id_t type );
+
+libspectrum_error
+internal_tzx_read( libspectrum_tape *tape, const libspectrum_byte *buffer,
+		   const size_t length );
+
+libspectrum_error
+internal_tzx_write( libspectrum_byte **buffer, size_t *length,
+		    libspectrum_tape *tape );
+
+libspectrum_error
+internal_warajevo_read( libspectrum_tape *tape,
+			const libspectrum_byte *buffer, size_t length );
+
+libspectrum_error
+libspectrum_z80em_read( libspectrum_tape *tape,
+                        const libspectrum_byte *buffer, size_t length );
+
+libspectrum_error
+libspectrum_csw_read( libspectrum_tape *tape,
+                      const libspectrum_byte *buffer, size_t length );
+
+libspectrum_error
+libspectrum_csw_write( libspectrum_byte **buffer, size_t *length,
+                       libspectrum_tape *tape );
+
+libspectrum_error
+libspectrum_wav_read( libspectrum_tape *tape, const char *filename );
+
 /* Crypto functions */
 
 libspectrum_error
@@ -127,4 +206,14 @@ libspectrum_sign_data( libspectrum_byte **signature, size_t *signature_length,
 		       libspectrum_byte *data, size_t data_length,
 		       libspectrum_rzx_dsa_key *key );
 
+libspectrum_tape_block*
+libspectrum_tape_block_internal_init(
+                                libspectrum_tape_block_state *iterator,
+                                libspectrum_tape *tape );
+
+libspectrum_error
+libspectrum_tape_get_next_edge_internal( libspectrum_dword *tstates, int *flags,
+                                         libspectrum_tape *tape,
+                                         libspectrum_tape_block_state *it );
+
 #endif				/* #ifndef LIBSPECTRUM_INTERNALS_H */
