@@ -1,7 +1,7 @@
 /* libspectrum.c: Some general routines
-   Copyright (c) 2001-2003 Philip Kendall, Darren Salt
+   Copyright (c) 2001-2004 Philip Kendall, Darren Salt, Fredrick Meunier
 
-   $Id: libspectrum.c,v 1.68.2.2 2004/01/29 23:17:34 pak21 Exp $
+   $Id: libspectrum.c 3439 2007-12-19 03:28:36Z zubzero $
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -13,14 +13,13 @@
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
 
-   You should have received a copy of the GNU General Public License
-   along with this program; if not, write to the Free Software
-   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
+   You should have received a copy of the GNU General Public License along
+   with this program; if not, write to the Free Software Foundation, Inc.,
+   51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 
    Author contact information:
 
-   E-mail: pak21-fuse@srcf.ucam.org
-   Postal address: 15 Crescent Road, Wokingham, Berks, RG40 2DB, England
+   E-mail: philip-fuse@shadowmagic.org.uk
 
 */
 
@@ -45,12 +44,35 @@ static const char *gcrypt_version;
 
 #include "internals.h"
 
+#if defined AMIGA || defined __MORPHOS__
+#include <proto/exec.h>
+#include <proto/xfdmaster.h>
+struct xfdMasterBase *xfdMasterBase;
+
+#ifndef __MORPHOS__
+struct xfdMasterIFace *IxfdMaster;
+#endif				/* #ifndef __MORPHOS__ */
+
+#endif /* #if defined AMIGA || defined __MORPHOS__ */
+
+/* The various real devices that can be attached to a virtual joystick */
+const int LIBSPECTRUM_JOYSTICK_INPUT_NONE             = 0;
+const int LIBSPECTRUM_JOYSTICK_INPUT_KEYBOARD         = 1 << 0;
+const int LIBSPECTRUM_JOYSTICK_INPUT_JOYSTICK_1       = 1 << 1;
+const int LIBSPECTRUM_JOYSTICK_INPUT_JOYSTICK_2       = 1 << 2;
+
+libspectrum_error
+libspectrum_default_error_function( libspectrum_error error,
+				    const char *format, va_list ap );
+
 /* The function to call on errors */
 libspectrum_error_function_t libspectrum_error_function =
   libspectrum_default_error_function;
 
+#ifdef HAVE_GCRYPT_H
 static void
 gcrypt_log_handler( void *opaque, int level, const char *format, va_list ap );
+#endif				/* #ifdef HAVE_GCRYPT_H */
 
 /* Initialise the library */
 libspectrum_error
@@ -94,11 +116,13 @@ libspectrum_init( void )
   return LIBSPECTRUM_ERROR_NONE;
 }
 
+#ifdef HAVE_GCRYPT_H
 static void
 gcrypt_log_handler( void *opaque, int level, const char *format, va_list ap )
 {
   /* Do nothing */
 }
+#endif				/* #ifdef HAVE_GCRYPT_H */
 
 int
 libspectrum_check_version( const char *version )
@@ -168,22 +192,50 @@ libspectrum_default_error_function( libspectrum_error error,
   return LIBSPECTRUM_ERROR_NONE;
 }
 
+/* Get the name of a specific joystick type */
+const char *
+libspectrum_joystick_name( libspectrum_joystick type )
+{
+  switch( type ) {
+  case LIBSPECTRUM_JOYSTICK_CURSOR:     return "Cursor";
+  case LIBSPECTRUM_JOYSTICK_KEMPSTON:   return "Kempston";
+  case LIBSPECTRUM_JOYSTICK_SINCLAIR_1: return "Sinclair 1";
+  case LIBSPECTRUM_JOYSTICK_SINCLAIR_2: return "Sinclair 2";
+  case LIBSPECTRUM_JOYSTICK_TIMEX_1:    return "Timex 1";
+  case LIBSPECTRUM_JOYSTICK_TIMEX_2:    return "Timex 2";
+  case LIBSPECTRUM_JOYSTICK_FULLER:     return "Fuller";
+
+  case LIBSPECTRUM_JOYSTICK_NONE:       return "(None)";
+  }
+
+  return "(unknown type)";
+}
+
 /* Get the name of a specific machine type */
 const char *
 libspectrum_machine_name( libspectrum_machine type )
 {
   switch( type ) {
-  case LIBSPECTRUM_MACHINE_16:     return "Spectrum 16K";
-  case LIBSPECTRUM_MACHINE_48:     return "Spectrum 48K";
-  case LIBSPECTRUM_MACHINE_TC2048: return "Timex TC2048";
-  case LIBSPECTRUM_MACHINE_TC2068: return "Timex TC2068";
-  case LIBSPECTRUM_MACHINE_128:    return "Spectrum 128K";
-  case LIBSPECTRUM_MACHINE_PLUS2:  return "Spectrum +2";
-  case LIBSPECTRUM_MACHINE_PENT:   return "Pentagon 128K";
-  case LIBSPECTRUM_MACHINE_PLUS2A: return "Spectrum +2A";
-  case LIBSPECTRUM_MACHINE_PLUS3:  return "Spectrum +3";
-  default:			   return "unknown";
+  case LIBSPECTRUM_MACHINE_16:       return "Spectrum 16K";
+  case LIBSPECTRUM_MACHINE_48:       return "Spectrum 48K";
+  case LIBSPECTRUM_MACHINE_TC2048:   return "Timex TC2048";
+  case LIBSPECTRUM_MACHINE_TC2068:   return "Timex TC2068";
+  case LIBSPECTRUM_MACHINE_TS2068:   return "Timex TS2068";
+  case LIBSPECTRUM_MACHINE_128:      return "Spectrum 128K";
+  case LIBSPECTRUM_MACHINE_PLUS2:    return "Spectrum +2";
+  case LIBSPECTRUM_MACHINE_PENT:     return "Pentagon 128K";
+  case LIBSPECTRUM_MACHINE_PENT512:  return "Pentagon 512K";
+  case LIBSPECTRUM_MACHINE_PENT1024: return "Pentagon 1024K";
+  case LIBSPECTRUM_MACHINE_PLUS2A:   return "Spectrum +2A";
+  case LIBSPECTRUM_MACHINE_PLUS3:    return "Spectrum +3";
+  case LIBSPECTRUM_MACHINE_PLUS3E:   return "Spectrum +3e";
+  case LIBSPECTRUM_MACHINE_SCORP:    return "Scorpion ZS 256";
+  case LIBSPECTRUM_MACHINE_SE:       return "Spectrum SE";
+
+  case LIBSPECTRUM_MACHINE_UNKNOWN: return "(unknown)";
   }
+
+  return "(unknown type)";
 }
 
 /* The various capabilities of the different machines */
@@ -206,6 +258,18 @@ const int LIBSPECTRUM_MACHINE_CAPABILITY_SINCLAIR_JOYSTICK = 1 << 8;
                                             /* Sinclair-style joystick ports */
 const int LIBSPECTRUM_MACHINE_CAPABILITY_KEMPSTON_JOYSTICK = 1 << 9;
                                             /* Kempston-style joystick ports */
+const int LIBSPECTRUM_MACHINE_CAPABILITY_SCORP_MEMORY = 1 << 10;
+                                             /* Scorpion-style memory paging */
+const int LIBSPECTRUM_MACHINE_CAPABILITY_EVEN_M1 = 1 << 11;
+                             /* M1 cycles always start on even tstate counts */
+const int LIBSPECTRUM_MACHINE_CAPABILITY_SE_MEMORY = 1 << 12;
+                                                   /* SE-style memory paging */
+const int LIBSPECTRUM_MACHINE_CAPABILITY_NTSC = 1 << 13;
+                                                             /* NTSC display */
+const int LIBSPECTRUM_MACHINE_CAPABILITY_PENT512_MEMORY = 1 << 14;
+					 /* Pentagon 512-style memory paging */
+const int LIBSPECTRUM_MACHINE_CAPABILITY_PENT1024_MEMORY = 1 << 15;
+					/* Pentagon 1024-style memory paging */
 
 /* Given a machine type, what features does it have? */
 int
@@ -217,8 +281,12 @@ libspectrum_machine_capabilities( libspectrum_machine type )
   switch( type ) {
   case LIBSPECTRUM_MACHINE_128: case LIBSPECTRUM_MACHINE_PLUS2:
   case LIBSPECTRUM_MACHINE_PLUS2A: case LIBSPECTRUM_MACHINE_PLUS3:
-  case LIBSPECTRUM_MACHINE_TC2068:
+  case LIBSPECTRUM_MACHINE_PLUS3E:
+  case LIBSPECTRUM_MACHINE_TC2068: case LIBSPECTRUM_MACHINE_TS2068:
   case LIBSPECTRUM_MACHINE_PENT:
+  case LIBSPECTRUM_MACHINE_PENT512: case LIBSPECTRUM_MACHINE_PENT1024:
+  case LIBSPECTRUM_MACHINE_SCORP:
+  case LIBSPECTRUM_MACHINE_SE:
     capabilities |= LIBSPECTRUM_MACHINE_CAPABILITY_AY; break;
   default:
     break;
@@ -228,7 +296,12 @@ libspectrum_machine_capabilities( libspectrum_machine type )
   switch( type ) {
   case LIBSPECTRUM_MACHINE_128: case LIBSPECTRUM_MACHINE_PLUS2:
   case LIBSPECTRUM_MACHINE_PLUS2A: case LIBSPECTRUM_MACHINE_PLUS3:
+  case LIBSPECTRUM_MACHINE_PLUS3E:
   case LIBSPECTRUM_MACHINE_PENT:
+  case LIBSPECTRUM_MACHINE_PENT512: case LIBSPECTRUM_MACHINE_PENT1024:
+  case LIBSPECTRUM_MACHINE_SCORP:
+/* FIXME: SE needs to have this capability to be considered a 128k machine */
+  case LIBSPECTRUM_MACHINE_SE:
     capabilities |= LIBSPECTRUM_MACHINE_CAPABILITY_128_MEMORY; break;
   default:
     break;
@@ -237,6 +310,7 @@ libspectrum_machine_capabilities( libspectrum_machine type )
   /* +3 Spectrum-style 0x1ffd memory paging */
   switch( type ) {
   case LIBSPECTRUM_MACHINE_PLUS2A: case LIBSPECTRUM_MACHINE_PLUS3:
+  case LIBSPECTRUM_MACHINE_PLUS3E:
     capabilities |= LIBSPECTRUM_MACHINE_CAPABILITY_PLUS3_MEMORY; break;
   default:
     break;
@@ -244,25 +318,25 @@ libspectrum_machine_capabilities( libspectrum_machine type )
 
   /* +3 Spectrum-style disk */
   switch( type ) {
-  case LIBSPECTRUM_MACHINE_PLUS3:
+  case LIBSPECTRUM_MACHINE_PLUS3: case LIBSPECTRUM_MACHINE_PLUS3E:
     capabilities |= LIBSPECTRUM_MACHINE_CAPABILITY_PLUS3_DISK; break;
   default:
     break;
   }
 
-  /* TC20[46]8-style 0x00fd memory paging */
+  /* T[CS]20[46]8-style 0x00fd memory paging */
   switch( type ) {
-  case LIBSPECTRUM_MACHINE_TC2048:
-  case LIBSPECTRUM_MACHINE_TC2068:
+  case LIBSPECTRUM_MACHINE_TC2048: case LIBSPECTRUM_MACHINE_TC2068:
+  case LIBSPECTRUM_MACHINE_TS2068:
     capabilities |= LIBSPECTRUM_MACHINE_CAPABILITY_TIMEX_MEMORY; break;
   default:
     break;
   }
 
-  /* TC20[46]8-style 0x00ff video mode selection */
+  /* T[CS]20[46]8-style 0x00ff video mode selection */
   switch( type ) {
-  case LIBSPECTRUM_MACHINE_TC2048:
-  case LIBSPECTRUM_MACHINE_TC2068:
+  case LIBSPECTRUM_MACHINE_TC2048: case LIBSPECTRUM_MACHINE_TC2068:
+  case LIBSPECTRUM_MACHINE_TS2068: case LIBSPECTRUM_MACHINE_SE:
     capabilities |= LIBSPECTRUM_MACHINE_CAPABILITY_TIMEX_VIDEO; break;
   default:
     break;
@@ -271,6 +345,8 @@ libspectrum_machine_capabilities( libspectrum_machine type )
   /* TRDOS-style disk */
   switch( type ) {
   case LIBSPECTRUM_MACHINE_PENT:
+  case LIBSPECTRUM_MACHINE_PENT512: case LIBSPECTRUM_MACHINE_PENT1024:
+  case LIBSPECTRUM_MACHINE_SCORP:
     capabilities |= LIBSPECTRUM_MACHINE_CAPABILITY_TRDOS_DISK; break;
   default:
     break;
@@ -278,7 +354,7 @@ libspectrum_machine_capabilities( libspectrum_machine type )
 
   /* T[SC]2068-style cartridge port */
   switch( type ) {
-  case LIBSPECTRUM_MACHINE_TC2068:
+  case LIBSPECTRUM_MACHINE_TC2068: case LIBSPECTRUM_MACHINE_TS2068:
     capabilities |= LIBSPECTRUM_MACHINE_CAPABILITY_TIMEX_DOCK; break;
   default:
     break;
@@ -287,7 +363,7 @@ libspectrum_machine_capabilities( libspectrum_machine type )
   /* Sinclair-style joystick ports */
   switch( type ) {
   case LIBSPECTRUM_MACHINE_PLUS2: case LIBSPECTRUM_MACHINE_PLUS2A:
-  case LIBSPECTRUM_MACHINE_PLUS3:
+  case LIBSPECTRUM_MACHINE_PLUS3: case LIBSPECTRUM_MACHINE_PLUS3E:
     capabilities |= LIBSPECTRUM_MACHINE_CAPABILITY_SINCLAIR_JOYSTICK; break;
   default:
     break;
@@ -296,11 +372,61 @@ libspectrum_machine_capabilities( libspectrum_machine type )
   /* Kempston-style joystick ports */
   switch( type ) {
   case LIBSPECTRUM_MACHINE_TC2048:
+  case LIBSPECTRUM_MACHINE_SE:
     capabilities |= LIBSPECTRUM_MACHINE_CAPABILITY_KEMPSTON_JOYSTICK; break;
   default:
     break;
   }
 
+  /* Scorpion-style 0x1ffd memory paging */
+  switch( type ) {
+  case LIBSPECTRUM_MACHINE_SCORP:
+    capabilities |= LIBSPECTRUM_MACHINE_CAPABILITY_SCORP_MEMORY; break;
+  default:
+    break;
+  }
+
+  /* M1 cycles forced to start on even tstates */
+  switch( type ) {
+  case LIBSPECTRUM_MACHINE_SCORP:
+    capabilities |= LIBSPECTRUM_MACHINE_CAPABILITY_EVEN_M1; break;
+  default:
+    break;
+  }
+
+  /* SE-style 0x7ffd and 0x00fd memory paging */
+  switch( type ) {
+  case LIBSPECTRUM_MACHINE_SE:
+    capabilities |= LIBSPECTRUM_MACHINE_CAPABILITY_SE_MEMORY; break;
+  default:
+    break;
+  }
+
+  /* NTSC display */
+  switch( type ) {
+  case LIBSPECTRUM_MACHINE_TS2068:
+    capabilities |= LIBSPECTRUM_MACHINE_CAPABILITY_NTSC; break;
+  default:
+    break;
+  }
+
+  /* Pentagon 512-style memory paging */
+  switch( type ) {
+  case LIBSPECTRUM_MACHINE_PENT512:
+  case LIBSPECTRUM_MACHINE_PENT1024:
+    capabilities |= LIBSPECTRUM_MACHINE_CAPABILITY_PENT512_MEMORY; break;
+  default:
+    break;
+  }
+
+  /* Pentagon 1024-style memory paging */
+  switch( type ) {
+  case LIBSPECTRUM_MACHINE_PENT1024:
+    capabilities |= LIBSPECTRUM_MACHINE_CAPABILITY_PENT1024_MEMORY; break;
+  default:
+    break;
+  }
+
   return capabilities;
 }
 
@@ -312,7 +438,7 @@ libspectrum_identify_file_with_class(
   const char *filename, const unsigned char *buffer, size_t length )
 {
   libspectrum_error error;
-  char *new_filename; unsigned char *new_buffer; size_t new_length;
+  char *new_filename = NULL; unsigned char *new_buffer; size_t new_length;
 
   error = libspectrum_identify_file_raw( type, filename, buffer, length );
   if( error ) return error;
@@ -368,6 +494,9 @@ libspectrum_identify_file_raw( libspectrum_id_t *type, const char *filename,
       { LIBSPECTRUM_ID_RECORDING_RZX, "rzx", 3, "RZX!",		    0, 4, 4 },
 
       { LIBSPECTRUM_ID_SNAPSHOT_SNA,  "sna", 3, NULL,		    0, 0, 0 },
+      /* Peter McGavin's Spectrum Emulator on the Amiga used .snapshot for sna
+         snaps */
+      { LIBSPECTRUM_ID_SNAPSHOT_SNA,  "snapshot", 3, NULL,	    0, 0, 0 },
       { LIBSPECTRUM_ID_SNAPSHOT_SNP,  "snp", 3, NULL,		    0, 0, 0 },
       { LIBSPECTRUM_ID_SNAPSHOT_SP,   "sp",  3, "\x53\x50\0",	    0, 3, 1 },
       { LIBSPECTRUM_ID_SNAPSHOT_SZX,  "szx", 3, "ZXST",		    0, 4, 4 },
@@ -375,10 +504,17 @@ libspectrum_identify_file_raw( libspectrum_id_t *type, const char *filename,
       /* .slt files also dealt with by the .z80 loading code */
       { LIBSPECTRUM_ID_SNAPSHOT_Z80,  "slt", 3, "\0\0",		    6, 2, 1 },
       { LIBSPECTRUM_ID_SNAPSHOT_ZXS,  "zxs", 3, "SNAP",		    8, 4, 4 },
+      { LIBSPECTRUM_ID_SNAPSHOT_PLUSD,"mgtsnp", 3, NULL,	    0, 0, 0 },
 
       { LIBSPECTRUM_ID_CARTRIDGE_DCK, "dck", 3, NULL,		    0, 0, 0 },
+      { LIBSPECTRUM_ID_CARTRIDGE_IF2, "rom", 3, NULL,		    0, 0, 0 },
+
+      { LIBSPECTRUM_ID_MICRODRIVE_MDR, "mdr", 3, NULL,		    0, 0, 0 },
 
       { LIBSPECTRUM_ID_TAPE_TAP,      "tap", 3, "\x13\0\0",	    0, 3, 1 },
+      { LIBSPECTRUM_ID_TAPE_SPC,      "spc", 3, "\x11\0\0",	    0, 3, 1 },
+      { LIBSPECTRUM_ID_TAPE_STA,      "sta", 3, "\x11\0\0",	    0, 3, 1 },
+      { LIBSPECTRUM_ID_TAPE_LTP,      "ltp", 3, "\x11\0\0",	    0, 3, 1 },
       { LIBSPECTRUM_ID_TAPE_TZX,      "tzx", 3, "ZXTape!",	    0, 7, 4 },
       { LIBSPECTRUM_ID_TAPE_WARAJEVO, "tap", 2, "\xff\xff\xff\xff", 8, 4, 2 },
 
@@ -386,9 +522,19 @@ libspectrum_identify_file_raw( libspectrum_id_t *type, const char *filename,
       { LIBSPECTRUM_ID_DISK_SCL,      "scl", 3, "SINCLAIR",         0, 8, 4 },
       { LIBSPECTRUM_ID_DISK_TRD,      "trd", 3, NULL,		    0, 0, 0 },
 
+      { LIBSPECTRUM_ID_HARDDISK_HDF,  "hdf", 3, "RS-IDE\x1a",	    0, 7, 4 },
+
       { LIBSPECTRUM_ID_COMPRESSED_BZ2,"bz2", 3, "BZh",		    0, 3, 4 },
       { LIBSPECTRUM_ID_COMPRESSED_GZ, "gz",  3, "\x1f\x8b",	    0, 2, 4 },
 
+      { LIBSPECTRUM_ID_TAPE_Z80EM,    "raw", 1, "\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0Raw tape sample",  0, 64, 0 },
+      { LIBSPECTRUM_ID_TAPE_CSW,      "csw", 2, "Compressed Square Wave\x1a",  0, 23, 4 },
+
+      { LIBSPECTRUM_ID_TAPE_WAV,      "wav", 3, NULL,		    0, 0, 0 },
+
+      { LIBSPECTRUM_ID_DISK_MGT,      "mgt", 3, NULL,		    0, 0, 0 },
+      { LIBSPECTRUM_ID_DISK_IMG,      "img", 3, NULL,		    0, 0, 0 },
+
       { -1, NULL, 0, NULL, 0, 0, 0 }, /* End marker */
 
     };
@@ -423,6 +569,56 @@ libspectrum_identify_file_raw( libspectrum_id_t *type, const char *filename,
     }
   }
 
+#if defined AMIGA || defined __MORPHOS__
+  /* Compressed file check through xfdmaster.library */
+
+  /* this prevents most good matches and gz/bz2 from being run through xfd (xfd
+     supports gz, but we want to use the internal code) */
+  if( best_score <= 3 ) {
+#ifndef __MORPHOS__
+    struct ExecIFace *IExec =
+      (struct ExecIFace *)(*(struct ExecBase **)4)->MainInterface;
+#endif				/* #ifndef __MORPHOS__ */
+    struct xfdBufferInfo *xfdobj;
+
+#ifndef __MORPHOS__
+    if( xfdMasterBase = IExec->OpenLibrary("xfdmaster.library",38 ) ) {
+      if( IxfdMaster =
+         (struct xfdMasterIFace *)IExec->GetInterface(xfdMasterBase,"main",1,NULL)){
+        if( xfdobj = (struct xfdBufferInfo *)IxfdMaster->xfdAllocObject(XFDOBJ_BUFFERINFO) ) {
+#else				/* #ifndef __MORPHOS__ */
+    if( xfdMasterBase = OpenLibrary("xfdmaster.library",38 ) ) {
+        if( xfdobj = (struct xfdBufferInfo *)xfdAllocObject(XFDOBJ_BUFFERINFO) ) {
+#endif				/* #ifndef __MORPHOS__ */
+          xfdobj->xfdbi_SourceBuffer = buffer;
+          xfdobj->xfdbi_SourceBufLen = length;
+          xfdobj->xfdbi_Flags = XFDFB_RECOGTARGETLEN | XFDFB_RECOGEXTERN;
+
+#ifndef __MORPHOS__
+          if( IxfdMaster->xfdRecogBuffer( xfdobj ) ) {
+#else				/* #ifndef __MORPHOS__ */
+          if( xfdRecogBuffer( xfdobj ) ) {
+#endif				/* #ifndef __MORPHOS__ */
+            best_guess = LIBSPECTRUM_ID_COMPRESSED_XFD;
+            duplicate_best=0;
+          }
+#ifndef __MORPHOS__
+          IxfdMaster->xfdFreeObject( (APTR)xfdobj );
+#else				/* #ifndef __MORPHOS__ */
+          xfdFreeObject( (APTR)xfdobj );
+#endif				/* #ifndef __MORPHOS__ */
+        }
+#ifndef __MORPHOS__
+        IExec->DropInterface( (struct Interface *)IxfdMaster );
+      }
+      IExec->CloseLibrary( xfdMasterBase );
+#else				/* #ifndef __MORPHOS__ */
+      CloseLibrary( xfdMasterBase );
+#endif				/* #ifndef __MORPHOS__ */
+    }
+  }
+#endif /* #ifdef AMIGA */
+
   /* If two things were equally good, we can't identify this. Otherwise,
      return our best guess */
   if( duplicate_best ) {
@@ -449,15 +645,26 @@ libspectrum_identify_class( libspectrum_class_t *libspectrum_class,
 
   case LIBSPECTRUM_ID_COMPRESSED_BZ2:
   case LIBSPECTRUM_ID_COMPRESSED_GZ:
+  case LIBSPECTRUM_ID_COMPRESSED_XFD:
     *libspectrum_class = LIBSPECTRUM_CLASS_COMPRESSED; return 0;
 
   case LIBSPECTRUM_ID_DISK_DSK:
     *libspectrum_class = LIBSPECTRUM_CLASS_DISK_PLUS3; return 0;
 
+  case LIBSPECTRUM_ID_DISK_IMG:
+  case LIBSPECTRUM_ID_DISK_MGT:
+    *libspectrum_class = LIBSPECTRUM_CLASS_DISK_PLUSD; return 0;
+
   case LIBSPECTRUM_ID_DISK_SCL:
   case LIBSPECTRUM_ID_DISK_TRD:
     *libspectrum_class = LIBSPECTRUM_CLASS_DISK_TRDOS; return 0;
 
+  case LIBSPECTRUM_ID_HARDDISK_HDF:
+    *libspectrum_class = LIBSPECTRUM_CLASS_HARDDISK; return 0;
+
+  case LIBSPECTRUM_ID_MICRODRIVE_MDR:
+    *libspectrum_class = LIBSPECTRUM_CLASS_MICRODRIVE; return 0;
+
   case LIBSPECTRUM_ID_RECORDING_RZX:
     *libspectrum_class = LIBSPECTRUM_CLASS_RECORDING; return 0;
 
@@ -471,10 +678,19 @@ libspectrum_identify_class( libspectrum_class_t *libspectrum_class,
     *libspectrum_class = LIBSPECTRUM_CLASS_SNAPSHOT; return 0;
 
   case LIBSPECTRUM_ID_TAPE_TAP:
+  case LIBSPECTRUM_ID_TAPE_SPC:
+  case LIBSPECTRUM_ID_TAPE_STA:
+  case LIBSPECTRUM_ID_TAPE_LTP:
   case LIBSPECTRUM_ID_TAPE_TZX:
   case LIBSPECTRUM_ID_TAPE_WARAJEVO:
+  case LIBSPECTRUM_ID_TAPE_Z80EM:
+  case LIBSPECTRUM_ID_TAPE_CSW:
+  case LIBSPECTRUM_ID_TAPE_WAV:
     *libspectrum_class = LIBSPECTRUM_CLASS_TAPE; return 0;
 
+  case LIBSPECTRUM_ID_CARTRIDGE_IF2:
+    *libspectrum_class = LIBSPECTRUM_CLASS_CARTRIDGE_IF2; return 0;
+
   }
 
   libspectrum_print_error( LIBSPECTRUM_ERROR_UNKNOWN,
@@ -518,7 +734,7 @@ libspectrum_uncompress_file( unsigned char **new_buffer, size_t *new_length,
 
 #ifdef HAVE_LIBBZ2
 
-    if( new_filename ) {
+    if( new_filename && *new_filename ) {
       if( strlen( *new_filename ) >= 4 &&
 	  !strcasecmp( &(*new_filename)[ strlen( *new_filename ) - 4 ],
 		       ".bz2" ) )
@@ -527,7 +743,10 @@ libspectrum_uncompress_file( unsigned char **new_buffer, size_t *new_length,
 
     error = libspectrum_bzip2_inflate( old_buffer, old_length,
 				       new_buffer, new_length );
-    if( error ) { free( new_filename ); return error; }
+    if( error ) {
+      if( new_filename ) free( *new_filename );
+      return error;
+    }
 
 #else				/* #ifdef HAVE_LIBBZ2 */
 
@@ -535,7 +754,7 @@ libspectrum_uncompress_file( unsigned char **new_buffer, size_t *new_length,
       LIBSPECTRUM_ERROR_UNKNOWN,
       "libbz2 not available to decompress bzipped file"
     );
-    free( new_filename );
+    if( new_filename ) free( *new_filename );
     return LIBSPECTRUM_ERROR_UNKNOWN;
 
 #endif				/* #ifdef HAVE_LIBBZ2 */
@@ -546,7 +765,7 @@ libspectrum_uncompress_file( unsigned char **new_buffer, size_t *new_length,
 
 #ifdef HAVE_ZLIB_H
 
-    if( new_filename ) {
+    if( new_filename && *new_filename ) {
       if( strlen( *new_filename ) >= 3 &&
 	  !strcasecmp( &(*new_filename)[ strlen( *new_filename ) - 3 ],
 		       ".gz" ) )
@@ -555,23 +774,99 @@ libspectrum_uncompress_file( unsigned char **new_buffer, size_t *new_length,
       
     error = libspectrum_gzip_inflate( old_buffer, old_length,
 				      new_buffer, new_length );
-    if( error ) { free( new_filename ); return error; }
+    if( error ) {
+      if( new_filename ) free( *new_filename );
+      return error;
+    }
 
 #else				/* #ifdef HAVE_ZLIB_H */
 
     libspectrum_print_error( LIBSPECTRUM_ERROR_UNKNOWN,
 			     "zlib not available to decompress gzipped file" );
-    free( new_filename );
+    if( new_filename ) free( *new_filename );
     return LIBSPECTRUM_ERROR_UNKNOWN;
 
 #endif				/* #ifdef HAVE_ZLIB_H */
 
     break;
 
+#if defined AMIGA || defined __MORPHOS__
+  case LIBSPECTRUM_ID_COMPRESSED_XFD:
+    {
+#ifndef __MORPHOS__
+      struct ExecIFace *IExec =
+        (struct ExecIFace *)(*(struct ExecBase **)4)->MainInterface;
+#endif				/* #ifndef __MORPHOS__ */
+      struct xfdBufferInfo *xfdobj;
+
+#ifndef __MORPHOS__
+      if( xfdMasterBase = IExec->OpenLibrary( "xfdmaster.library", 38 ) ) {
+        if( IxfdMaster =
+            (struct xfdMasterIFace *)IExec->GetInterface(xfdMasterBase,"main",1,NULL)) {
+          if( xfdobj =
+              (struct xfdBufferInfo *)IxfdMaster->xfdAllocObject(XFDOBJ_BUFFERINFO) ) {
+#else				/* #ifndef __MORPHOS__ */
+      if( xfdMasterBase = OpenLibrary( "xfdmaster.library", 38 ) ) {
+          if( xfdobj = (struct xfdBufferInfo *)xfdAllocObject(XFDOBJ_BUFFERINFO) ) {
+#endif				/* #ifndef __MORPHOS__ */
+            xfdobj->xfdbi_SourceBufLen = old_length;
+            xfdobj->xfdbi_SourceBuffer = old_buffer;
+            xfdobj->xfdbi_Flags = XFDFB_RECOGEXTERN | XFDFB_RECOGTARGETLEN;
+            xfdobj->xfdbi_PackerFlags = XFDPFB_RECOGLEN;
+#ifndef __MORPHOS__
+            if( IxfdMaster->xfdRecogBuffer( xfdobj ) ) {
+#else				/* #ifndef __MORPHOS__ */
+            if( xfdRecogBuffer( xfdobj ) ) {
+#endif				/* #ifndef __MORPHOS__ */
+              xfdobj->xfdbi_TargetBufMemType = MEMF_ANY;
+#ifndef __MORPHOS__
+              if( IxfdMaster->xfdDecrunchBuffer( xfdobj ) ) {
+#else				/* #ifndef __MORPHOS__ */
+              if( xfdDecrunchBuffer( xfdobj ) ) {
+#endif				/* #ifndef __MORPHOS__ */
+                *new_buffer = malloc( xfdobj->xfdbi_TargetBufSaveLen );
+                *new_length = xfdobj->xfdbi_TargetBufSaveLen;
+                memcpy( *new_buffer, xfdobj->xfdbi_TargetBuffer, *new_length );
+#ifndef __MORPHOS__
+                IExec->FreeMem( xfdobj->xfdbi_TargetBuffer,xfdobj->xfdbi_TargetBufLen );
+#else				/* #ifndef __MORPHOS__ */
+                FreeMem( xfdobj->xfdbi_TargetBuffer,xfdobj->xfdbi_TargetBufLen );
+#endif				/* #ifndef __MORPHOS__ */
+              } else {
+                libspectrum_print_error(
+                             LIBSPECTRUM_ERROR_UNKNOWN,
+                             "xfdmaster.library not able to decrunch %s file",
+                             xfdobj->xfdbi_PackerName );
+                return LIBSPECTRUM_ERROR_UNKNOWN;
+              }
+            } else {
+              libspectrum_print_error(
+                                 LIBSPECTRUM_ERROR_UNKNOWN,
+                                 "xfdmaster.library does not recognise file" );
+              return LIBSPECTRUM_ERROR_UNKNOWN;
+            }
+#ifndef __MORPHOS__
+            IxfdMaster->xfdFreeObject( xfdobj );
+#else				/* #ifndef __MORPHOS__ */
+            xfdFreeObject( xfdobj );
+#endif				/* #ifndef __MORPHOS__ */
+          }
+#ifndef __MORPHOS__
+          IExec->DropInterface( (struct Interface *)IxfdMaster );
+        }
+        IExec->CloseLibrary( xfdMasterBase );
+#else				/* #ifndef __MORPHOS__ */
+        CloseLibrary( xfdMasterBase );
+#endif				/* #ifndef __MORPHOS__ */
+      }
+    }
+    break;
+#endif /* #ifdef AMIGA */
+
   default:
     libspectrum_print_error( LIBSPECTRUM_ERROR_LOGIC,
 			     "unknown compressed type %d", type );
-    free( new_filename );
+    if( new_filename ) free( *new_filename );
     return LIBSPECTRUM_ERROR_LOGIC;
   }
 
