@@ -1,7 +1,7 @@
-/* make-sed.c: Generate a sed script to create the libspectrum_* typedefs
+/* make-perl.c: Generate a perl script to create the libspectrum_* typedefs
    Copyright (c) 2002-2003 Philip Kendall, Darren Salt
 
-   $Id: make-perl.c,v 1.11 2003/12/14 12:05:31 pak21 Exp $
+   $Id: make-perl.c 3023 2007-06-28 19:01:49Z zubzero $
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -13,14 +13,13 @@
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
 
-   You should have received a copy of the GNU General Public License
-   along with this program; if not, write to the Free Software
-   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
+   You should have received a copy of the GNU General Public License along
+   with this program; if not, write to the Free Software Foundation, Inc.,
+   51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 
    Author contact information:
 
-   E-mail: pak21-fuse@srcf.ucam.org
-   Postal address: 15 Crescent Road, Wokingham, Berks, RG40 2DB, England
+   E-mail: philip-fuse@shadowmagic.org.uk
 
 */
 
@@ -115,7 +114,17 @@ int main(void)
 #else				/* #ifdef HAVE_LIB_GLIB */
 
   printf( "  $_ = << \"CODE\";\n"
+"#ifndef	FALSE\n"
+"#define	FALSE	(0)\n"
+"#endif\n"
+"\n"
+"#ifndef	TRUE\n"
+"#define	TRUE	(!FALSE)\n"
+"#endif\n"
+"\n"
 "typedef int gint;\n"
+"typedef long glong;\n"
+"typedef gint gboolean;\n"
 "typedef unsigned int guint;\n"
 "typedef const void * gconstpointer;\n"
 "typedef void * gpointer;\n"
@@ -134,36 +143,85 @@ int main(void)
 "						 gconstpointer	b);\n"
 "\n"
 "\n"
-"GSList* g_slist_insert_sorted	(GSList		*list,\n"
-"				 gpointer	 data,\n"
-"				 GCompareFunc	 func);\n"
+"GSList WIN32_DLL *g_slist_insert_sorted	(GSList		*list,\n"
+"						 gpointer	 data,\n"
+"						 GCompareFunc	 func);\n"
 "\n"
-"GSList* g_slist_append		(GSList		*list,\n"
-"				 gpointer	 data);\n"
+"GSList WIN32_DLL *g_slist_insert		(GSList		*list,\n"
+"			 			 gpointer	 data,\n"
+"			 			 gint		 position);\n"
 "\n"
-"GSList* g_slist_prepend	(GSList		*list,\n"
-"				 gpointer	 data);\n"
+"GSList WIN32_DLL *g_slist_append		(GSList		*list,\n"
+"						 gpointer	 data);\n"
 "\n"
-"GSList* g_slist_remove		(GSList		*list,\n"
-"				 gpointer	 data);\n"
+"GSList WIN32_DLL *g_slist_prepend		(GSList		*list,\n"
+"						 gpointer	 data);\n"
 "\n"
-"void	g_slist_foreach		(GSList		*list,\n"
-"				 GFunc		 func,\n"
-"				 gpointer	 user_data);\n"
+"GSList WIN32_DLL *g_slist_remove		(GSList		*list,\n"
+"						 gpointer	 data);\n"
 "\n"
-"void	g_slist_free		(GSList		*list);\n"
+"GSList WIN32_DLL *g_slist_delete_link		(GSList		*list,\n"
+"				 		 GSList		*link);\n"
 "\n"
-"GSList* g_slist_nth		(GSList		*list,\n"
-"				 guint		n);\n"
+"guint	WIN32_DLL g_slist_length		(GSList *list);\n"
 "\n"
-"GSList* g_slist_find_custom	(GSList		*list,\n"
-"				 gpointer	data,\n"
-"				 GCompareFunc	func );\n"
+"void	WIN32_DLL g_slist_foreach		(GSList		*list,\n"
+"						 GFunc		 func,\n"
+"						 gpointer	 user_data);\n"
 "\n"
-"gint	g_slist_position	(GSList		*list,\n"
-"				 GSList		*llink);\n"
-"CODE\n" );
-
+"void	WIN32_DLL g_slist_free			(GSList		*list);\n"
+"\n"
+"GSList WIN32_DLL *g_slist_nth		(GSList		*list,\n"
+"					 guint		n);\n"
+"\n"
+"GSList WIN32_DLL * g_slist_find_custom	(GSList		*list,\n"
+"					 gpointer	data,\n"
+"					 GCompareFunc	func );\n"
+"\n"
+"gint	WIN32_DLL g_slist_position	(GSList		*list,\n"
+"					 GSList		*llink);\n"
+"\n"
+"typedef struct _GHashTable	GHashTable;\n"
+"\n"
+"typedef guint		(*GHashFunc)		(gconstpointer	key);\n"
+"\n"
+"typedef gboolean	(*GHRFunc)		(gpointer	key,\n"
+"						 gpointer	value,\n"
+"						 gpointer	user_data);\n"
+"\n"
+"gint	WIN32_DLL g_int_equal (gconstpointer   v,\n"
+"			       gconstpointer   v2);\n"
+"guint	WIN32_DLL g_int_hash  (gconstpointer   v);\n"
+"\n"
+"GHashTable WIN32_DLL *g_hash_table_new	(GHashFunc	 hash_func,\n"
+"					 GCompareFunc	 key_compare_func);\n"
+"\n"
+"void	WIN32_DLL g_hash_table_destroy	(GHashTable	*hash_table);\n"
+"\n"
+"void	WIN32_DLL g_hash_table_insert	(GHashTable	*hash_table,\n"
+"					 gpointer	 key,\n"
+"					 gpointer	 value);\n"
+"\n"
+"gpointer WIN32_DLL g_hash_table_lookup	(GHashTable	*hash_table,\n"
+"					 gconstpointer	 key);\n"
+"\n"
+"guint	WIN32_DLL g_hash_table_foreach_remove	(GHashTable	*hash_table,\n"
+"						 GHRFunc	 func,\n"
+"						 gpointer	 user_data);\n"
+"\n"
+"guint	WIN32_DLL g_hash_table_size (GHashTable	*hash_table);\n"
+"\n" );
+  if( sizeof( void* ) == sizeof( int ) ) {
+    printf( "#define GINT_TO_POINTER(i)      ((gpointer)  (i))\n" );
+    printf( "#define GPOINTER_TO_INT(p)      ((gint)   (p))\n" );
+  } else if( sizeof( void* ) == sizeof( long ) ) {
+    printf( "#define GINT_TO_POINTER(i)      ((gpointer)  (glong)(i))\n" );
+    printf( "#define GPOINTER_TO_INT(p)      ((gint)   (glong)(p))\n" );
+  } else {
+    fprintf( stderr, "No plausible int to pointer cast found\n" );
+    return 1;
+  }
+  printf( "CODE\n" );
 #endif				/* #ifdef HAVE_LIB_GLIB */
 
   printf( "}\n\n" );
@@ -201,5 +259,16 @@ int main(void)
 
   printf( "}\n\n" );
 
+  printf( "if( /LIBSPECTRUM_CAPABILITIES/ ) {\n\n  $_ = << \"CODE\";\n" );
+
+#ifdef HAVE_ZLIB_H
+
+  printf( "\n/* we support snapshots etc. requiring zlib (e.g. compressed szx) */\n" );
+  printf( "#define	LIBSPECTRUM_SUPPORTS_ZLIB_COMPRESSION	(1)\n" );
+
+#endif				/* #ifdef HAVE_ZLIB_H */
+
+  printf( "CODE\n}\n\n" );
+
   return 0;
 }
