@@ -1,7 +1,7 @@
 /* gslist.c: Minimal replacement for GSList
    Copyright (c) 2001-2004 Matan Ziv-Av, Philip Kendall, Marek Januszewski
 
-   $Id: gslist.c,v 1.18 2004/04/14 15:31:09 pak21 Exp $
+   $Id: gslist.c 2890 2007-05-26 19:31:43Z zubzero $
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -13,14 +13,13 @@
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
 
-   You should have received a copy of the GNU General Public License
-   along with this program; if not, write to the Free Software
-   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
+   You should have received a copy of the GNU General Public License along
+   with this program; if not, write to the Free Software Foundation, Inc.,
+   51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 
    Author contact information:
 
-   E-mail: pak21-fuse@srcf.ucam.org
-   Postal address: 15 Crescent Road, Wokingham, Berks, RG40 2DB, England
+   E-mail: philip-fuse@shadowmagic.org.uk
 
 */
 
@@ -43,6 +42,61 @@ gint	last_function		(gconstpointer	 a,
 
 GSList * free_list = NULL;
 
+GSList* g_slist_insert	(GSList		*list,
+			 gpointer	 data,
+			 gint		 position) {
+    
+  GSList *prev_list;
+  GSList *tmp_list;
+  GSList *new_list;
+
+  if (position < 0)
+    return g_slist_append (list, data);
+  else if (position == 0)
+    return g_slist_prepend (list, data);
+
+
+    if(!free_list) {
+        int i;
+        free_list=(GSList *)malloc(1024*sizeof(GSList));
+        for(i=0;i<1023;i++)
+            free_list[i].next=&free_list[i+1];
+        free_list[1023].next=NULL;
+    }
+
+  new_list = free_list;
+  free_list=free_list->next;
+  new_list->data = data;
+  new_list->next=NULL;
+
+  if (!list)
+    {
+      return new_list;
+    }
+
+  prev_list = NULL;
+  tmp_list = list;
+
+  while ((position-- > 0) && tmp_list)
+    {
+      prev_list = tmp_list;
+      tmp_list = tmp_list->next;
+    }
+
+  if (prev_list)
+    {
+      new_list->next = prev_list->next;
+      prev_list->next = new_list;
+    }
+  else
+    {
+      new_list->next = list;
+      list = new_list;
+    }
+
+  return list;
+}
+
 GSList* g_slist_insert_sorted	(GSList		*list,
 				 gpointer	 data,
 				 GCompareFunc	 func) {
@@ -162,6 +216,51 @@ GSList* g_slist_remove		(GSList		*list,
   return list;
 }
 
+GSList* g_slist_delete_link	(GSList		*list,
+				 GSList		*link) {
+
+  GSList *tmp;
+  GSList *prev;
+
+  prev = NULL;
+  tmp = list;
+
+  while (tmp)
+    {
+      if (tmp == link)
+        {
+          if (prev)
+            prev->next = tmp->next;
+          if (list == tmp)
+            list = list->next;
+
+          tmp->next = NULL;
+          g_slist_free (tmp);
+          break;
+        }
+
+      prev = tmp;
+      tmp = tmp->next;
+    }
+
+  return list;
+}
+
+guint
+g_slist_length (GSList *list)
+{
+  guint length;
+
+  length = 0;
+  while (list)
+    {
+      length++;
+      list = list->next;
+    }
+
+  return length;
+}
+
 void	g_slist_foreach		(GSList		*list,
 				 GFunc		 func,
 				 gpointer	 user_data) {
