@@ -1,7 +1,7 @@
 /* sna.c: Routines for handling .sna snapshots
    Copyright (c) 2001-2002 Philip Kendall
 
-   $Id: sna.c,v 1.23 2004/06/02 13:38:10 pak21 Exp $
+   $Id: sna.c 3351 2007-11-29 15:59:47Z pak21 $
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -13,14 +13,13 @@
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
 
-   You should have received a copy of the GNU General Public License
-   along with this program; if not, write to the Free Software
-   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
+   You should have received a copy of the GNU General Public License along
+   with this program; if not, write to the Free Software Foundation, Inc.,
+   51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 
    Author contact information:
 
-   E-mail: pak21-fuse@srcf.ucam.org
-   Postal address: 15 Crescent Road, Wokingham, Berks, RG40 2DB, England
+   E-mail: philip-fuse@shadowmagic.org.uk
 
 */
 
@@ -62,6 +61,13 @@ write_page( libspectrum_byte *buffer, libspectrum_snap *snap, int page );
 libspectrum_error
 libspectrum_sna_read( libspectrum_snap *snap,
 	              const libspectrum_byte *buffer, size_t buffer_length )
+{
+  return internal_sna_read( snap, buffer, buffer_length );
+}
+
+libspectrum_error
+internal_sna_read( libspectrum_snap *snap,
+		   const libspectrum_byte *buffer, size_t buffer_length )
 {
   int error;
 
@@ -144,9 +150,8 @@ static int
 libspectrum_sna_read_data( const libspectrum_byte *buffer,
 			   size_t buffer_length, libspectrum_snap *snap )
 {
-  int error;
-  int offset; int page;
-  int i,j;
+  int error, page, i, j;
+  libspectrum_word sp, offset;
 
   if( buffer_length < 0xc000 ) {
     libspectrum_print_error(
@@ -160,8 +165,17 @@ libspectrum_sna_read_data( const libspectrum_byte *buffer,
 
   case LIBSPECTRUM_MACHINE_48:
 
+    sp = libspectrum_snap_sp( snap );
+    if( sp < 0x4000 || sp == 0xffff ) {
+      libspectrum_print_error(
+        LIBSPECTRUM_ERROR_CORRUPT,
+        "libspectrum_sna_read_data: SP invalid (0x%04x)", sp
+      );
+      return LIBSPECTRUM_ERROR_CORRUPT;
+    }
+
     /* Rescue PC from the stack */
-    offset = libspectrum_snap_sp( snap ) - 0x4000;
+    offset = sp - 0x4000;
     libspectrum_snap_set_pc( snap, buffer[offset] + 0x100 * buffer[offset+1] );
 
     /* Increase SP as PC has been unstacked */
@@ -314,6 +328,7 @@ libspectrum_sna_write( libspectrum_byte **buffer, size_t *length,
 
   case LIBSPECTRUM_MACHINE_TC2048:
   case LIBSPECTRUM_MACHINE_TC2068:
+  case LIBSPECTRUM_MACHINE_TS2068:
     *out_flags |= LIBSPECTRUM_FLAG_SNAPSHOT_MAJOR_INFO_LOSS;
     /* Fall through */
   case LIBSPECTRUM_MACHINE_16:
@@ -325,6 +340,8 @@ libspectrum_sna_write( libspectrum_byte **buffer, size_t *length,
   case LIBSPECTRUM_MACHINE_PLUS3:
   case LIBSPECTRUM_MACHINE_PLUS3E:
   case LIBSPECTRUM_MACHINE_PENT:
+  case LIBSPECTRUM_MACHINE_PENT512:
+  case LIBSPECTRUM_MACHINE_PENT1024:
   case LIBSPECTRUM_MACHINE_SCORP:
   case LIBSPECTRUM_MACHINE_SE:
     *out_flags |= LIBSPECTRUM_FLAG_SNAPSHOT_MAJOR_INFO_LOSS;
@@ -411,7 +428,7 @@ write_48k_sna( libspectrum_byte **buffer, libspectrum_byte **ptr,
   error = write_page( &( (*ptr)[ 0x8000 ] ), snap, 0 );
   if( error ) return error;
 
-  /* Overwrite a bit of memory with SP */
+  /* Place PC on the stack */
   stack = &( (*ptr)[ libspectrum_snap_sp( snap ) - 0x4000 - 2 ] );
   libspectrum_write_word( &stack, libspectrum_snap_pc( snap ) );
 
@@ -476,4 +493,3 @@ write_page( libspectrum_byte *buffer, libspectrum_snap *snap, int page )
 
   return LIBSPECTRUM_ERROR_NONE;
 }
-
