@@ -1,7 +1,7 @@
 /* sp.c: Routines for handling .sp snapshots
    Copyright (c) 1998,2003 Philip Kendall
 
-   $Id: sp.c,v 1.3 2003/08/14 13:28:16 pak21 Exp $
+   $Id: sp.c 2890 2007-05-26 19:31:43Z zubzero $
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -13,14 +13,13 @@
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
 
-   You should have received a copy of the GNU General Public License
-   along with this program; if not, write to the Free Software
-   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
+   You should have received a copy of the GNU General Public License along
+   with this program; if not, write to the Free Software Foundation, Inc.,
+   51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 
    Author contact information:
 
-   E-mail: pak21-fuse@srcf.ucam.org
-   Postal address: 15 Crescent Road, Wokingham, Berks, RG40 2DB, England
+   E-mail: philip-fuse@shadowmagic.org.uk
 
 */
 
@@ -30,7 +29,7 @@
 
 #include "internals.h"
 
-static const int SP_HEADER_LENGTH = 37;
+static const size_t SP_HEADER_LENGTH = 37;
 
 libspectrum_error
 libspectrum_sp_read( libspectrum_snap *snap, const libspectrum_byte *buffer,
