@@ -1,7 +1,7 @@
 /* tzx_read.c: Routines for reading .tzx files
-   Copyright (c) 2001, 2002 Philip Kendall, Darren Salt
+   Copyright (c) 2001-2007 Philip Kendall, Darren Salt
 
-   $Id: tzx_read.c,v 1.52.2.1 2004/02/11 21:20:49 pak21 Exp $
+   $Id: tzx_read.c 3071 2007-07-20 17:12:49Z pak21 $
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -13,19 +13,19 @@
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
 
-   You should have received a copy of the GNU General Public License
-   along with this program; if not, write to the Free Software
-   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
+   You should have received a copy of the GNU General Public License along
+   with this program; if not, write to the Free Software Foundation, Inc.,
+   51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 
    Author contact information:
 
-   E-mail: pak21-fuse@srcf.ucam.org
-   Postal address: 15 Crescent Road, Wokingham, Berks, RG40 2DB, England
+   E-mail: philip-fuse@shadowmagic.org.uk
 
 */
 
 #include <config.h>
 
+#include <math.h>
 #include <stddef.h>
 #include <stdio.h>
 #include <string.h>
@@ -33,7 +33,7 @@
 #include "internals.h"
 
 /* The .tzx file signature (first 8 bytes) */
-static const libspectrum_byte *signature = "ZXTape!\x1a";
+const char *libspectrum_tzx_signature = "ZXTape!\x1a";
 
 /*** Local function prototypes ***/
 
@@ -56,6 +56,10 @@ static libspectrum_error
 tzx_read_raw_data( libspectrum_tape *tape, const libspectrum_byte **ptr,
 		   const libspectrum_byte *end );
 static libspectrum_error
+tzx_read_generalised_data( libspectrum_tape *tape,
+			   const libspectrum_byte **ptr,
+			   const libspectrum_byte *end );
+static libspectrum_error
 tzx_read_pause( libspectrum_tape *tape, const libspectrum_byte **ptr,
 		const libspectrum_byte *end );
 static libspectrum_error
@@ -99,26 +103,27 @@ tzx_read_data( const libspectrum_byte **ptr, const libspectrum_byte *end,
 	       size_t *length, int bytes, libspectrum_byte **data );
 static libspectrum_error
 tzx_read_string( const libspectrum_byte **ptr, const libspectrum_byte *end,
-		 libspectrum_byte **dest );
+		 char **dest );
 
 /*** Function definitions ***/
 
 /* The main load function */
 
 libspectrum_error
-libspectrum_tzx_read( libspectrum_tape *tape, const libspectrum_byte *buffer,
-		      const size_t length )
+internal_tzx_read( libspectrum_tape *tape, const libspectrum_byte *buffer,
+		   const size_t length )
 {
 
   libspectrum_error error;
 
   const libspectrum_byte *ptr, *end;
+  size_t signature_length = strlen( libspectrum_tzx_signature );
 
   ptr = buffer; end = buffer + length;
 
   /* Must be at least as many bytes as the signature, and the major/minor
      version numbers */
-  if( length < strlen(signature) + 2 ) {
+  if( length < signature_length + 2 ) {
     libspectrum_print_error(
       LIBSPECTRUM_ERROR_CORRUPT,
       "libspectrum_tzx_create: not enough data in buffer"
@@ -127,12 +132,12 @@ libspectrum_tzx_read( libspectrum_tape *tape, const libspectrum_byte *buffer,
   }
 
   /* Now check the signature */
-  if( memcmp( ptr, signature, strlen( signature ) ) ) {
+  if( memcmp( ptr, libspectrum_tzx_signature, signature_length ) ) {
     libspectrum_print_error( LIBSPECTRUM_ERROR_SIGNATURE,
 			     "libspectrum_tzx_create: wrong signature" );
     return LIBSPECTRUM_ERROR_SIGNATURE;
   }
-  ptr += strlen( signature );
+  ptr += signature_length;
   
   /* Just skip the version numbers */
   ptr += 2;
@@ -145,93 +150,98 @@ libspectrum_tzx_read( libspectrum_tape *tape, const libspectrum_byte *buffer,
     switch( id ) {
     case LIBSPECTRUM_TAPE_BLOCK_ROM:
       error = tzx_read_rom_block( tape, &ptr, end );
-       if( error ) { libspectrum_tape_free( tape ); return error; }
+      if( error ) { libspectrum_tape_clear( tape ); return error; }
       break;
     case LIBSPECTRUM_TAPE_BLOCK_TURBO:
       error = tzx_read_turbo_block( tape, &ptr, end );
-      if( error ) { libspectrum_tape_free( tape ); return error; }
+      if( error ) { libspectrum_tape_clear( tape ); return error; }
       break;
     case LIBSPECTRUM_TAPE_BLOCK_PURE_TONE:
       error = tzx_read_pure_tone( tape, &ptr, end );
-      if( error ) { libspectrum_tape_free( tape ); return error; }
+      if( error ) { libspectrum_tape_clear( tape ); return error; }
       break;
     case LIBSPECTRUM_TAPE_BLOCK_PULSES:
       error = tzx_read_pulses_block( tape, &ptr, end );
-      if( error ) { libspectrum_tape_free( tape ); return error; }
+      if( error ) { libspectrum_tape_clear( tape ); return error; }
       break;
     case LIBSPECTRUM_TAPE_BLOCK_PURE_DATA:
       error = tzx_read_pure_data( tape, &ptr, end );
-      if( error ) { libspectrum_tape_free( tape ); return error; }
+      if( error ) { libspectrum_tape_clear( tape ); return error; }
       break;
     case LIBSPECTRUM_TAPE_BLOCK_RAW_DATA:
       error = tzx_read_raw_data( tape, &ptr, end );
-      if( error ) { libspectrum_tape_free( tape ); return error; }
+      if( error ) { libspectrum_tape_clear( tape ); return error; }
+      break;
+
+    case LIBSPECTRUM_TAPE_BLOCK_GENERALISED_DATA:
+      error = tzx_read_generalised_data( tape, &ptr, end );
+      if( error ) { libspectrum_tape_clear( tape ); return error; }
       break;
 
     case LIBSPECTRUM_TAPE_BLOCK_PAUSE:
       error = tzx_read_pause( tape, &ptr, end );
-      if( error ) { libspectrum_tape_free( tape ); return error; }
+      if( error ) { libspectrum_tape_clear( tape ); return error; }
       break;
     case LIBSPECTRUM_TAPE_BLOCK_GROUP_START:
       error = tzx_read_group_start( tape, &ptr, end );
-      if( error ) { libspectrum_tape_free( tape ); return error; }
+      if( error ) { libspectrum_tape_clear( tape ); return error; }
       break;
     case LIBSPECTRUM_TAPE_BLOCK_GROUP_END:
       error = tzx_read_empty_block( tape, id );
-      if( error ) { libspectrum_tape_free( tape ); return error; }
+      if( error ) { libspectrum_tape_clear( tape ); return error; }
       break;
     case LIBSPECTRUM_TAPE_BLOCK_JUMP:
       error = tzx_read_jump( tape, &ptr, end );
-      if( error ) { libspectrum_tape_free( tape ); return error; }
+      if( error ) { libspectrum_tape_clear( tape ); return error; }
       break;
     case LIBSPECTRUM_TAPE_BLOCK_LOOP_START:
       error = tzx_read_loop_start( tape, &ptr, end );
-      if( error ) { libspectrum_tape_free( tape ); return error; }
+      if( error ) { libspectrum_tape_clear( tape ); return error; }
       break;
     case LIBSPECTRUM_TAPE_BLOCK_LOOP_END:
       error = tzx_read_empty_block( tape, id );
-      if( error ) { libspectrum_tape_free( tape ); return error; }
+      if( error ) { libspectrum_tape_clear( tape ); return error; }
       break;
 
     case LIBSPECTRUM_TAPE_BLOCK_SELECT:
       error = tzx_read_select( tape, &ptr, end );
-      if( error ) { libspectrum_tape_free( tape ); return error; }
+      if( error ) { libspectrum_tape_clear( tape ); return error; }
       break;
 
     case LIBSPECTRUM_TAPE_BLOCK_STOP48:
       error = tzx_read_stop( tape, &ptr, end );
-      if( error ) { libspectrum_tape_free( tape ); return error; }
+      if( error ) { libspectrum_tape_clear( tape ); return error; }
       break;
 
     case LIBSPECTRUM_TAPE_BLOCK_COMMENT:
       error = tzx_read_comment( tape, &ptr, end );
-      if( error ) { libspectrum_tape_free( tape ); return error; }
+      if( error ) { libspectrum_tape_clear( tape ); return error; }
       break;
     case LIBSPECTRUM_TAPE_BLOCK_MESSAGE:
       error = tzx_read_message( tape, &ptr, end );
-      if( error ) { libspectrum_tape_free( tape ); return error; }
+      if( error ) { libspectrum_tape_clear( tape ); return error; }
       break;
     case LIBSPECTRUM_TAPE_BLOCK_ARCHIVE_INFO:
       error = tzx_read_archive_info( tape, &ptr, end );
-      if( error ) { libspectrum_tape_free( tape ); return error; }
+      if( error ) { libspectrum_tape_clear( tape ); return error; }
       break;
     case LIBSPECTRUM_TAPE_BLOCK_HARDWARE:
       error = tzx_read_hardware( tape, &ptr, end );
-      if( error ) { libspectrum_tape_free( tape ); return error; }
+      if( error ) { libspectrum_tape_clear( tape ); return error; }
       break;
 
     case LIBSPECTRUM_TAPE_BLOCK_CUSTOM:
       error = tzx_read_custom( tape, &ptr, end );
-      if( error ) { libspectrum_tape_free( tape ); return error; }
+      if( error ) { libspectrum_tape_clear( tape ); return error; }
       break;
 
     case LIBSPECTRUM_TAPE_BLOCK_CONCAT:
       error = tzx_read_concat( &ptr, end );
-      if( error ) { libspectrum_tape_free( tape ); return error; }
+      if( error ) { libspectrum_tape_clear( tape ); return error; }
       break;
 
     default:	/* For now, don't handle anything else */
-      libspectrum_tape_free( tape );
+      libspectrum_tape_clear( tape );
       libspectrum_print_error(
         LIBSPECTRUM_ERROR_UNKNOWN,
         "libspectrum_tzx_create: unknown block type 0x%02x", id
@@ -243,6 +253,13 @@ libspectrum_tzx_read( libspectrum_tape *tape, const libspectrum_byte *buffer,
   return LIBSPECTRUM_ERROR_NONE;
 }
 
+libspectrum_error
+libspectrum_tzx_read( libspectrum_tape *tape, const libspectrum_byte *buffer,
+		      const size_t length )
+{
+  return internal_tzx_read( tape, buffer, length );
+}
+
 static libspectrum_error
 tzx_read_rom_block( libspectrum_tape *tape, const libspectrum_byte **ptr,
 		    const libspectrum_byte *end )
@@ -526,6 +543,162 @@ tzx_read_raw_data (libspectrum_tape *tape, const libspectrum_byte **ptr,
   return LIBSPECTRUM_ERROR_NONE;
 }
 
+static libspectrum_error
+tzx_read_generalised_data( libspectrum_tape *tape,
+			   const libspectrum_byte **ptr,
+			   const libspectrum_byte *end )
+{
+  libspectrum_tape_block *block;
+  libspectrum_dword length, symbol_count, data_count;
+  libspectrum_word symbol_count2;
+  libspectrum_error error;
+  libspectrum_tape_generalised_data_symbol_table *table;
+  libspectrum_byte *symbols, *data;
+  libspectrum_word *repeats;
+
+  const libspectrum_byte *blockend, *ptr2;
+  size_t i, data_size;
+  int bits_per_symbol;
+
+  /* Check the length exists */
+  if( end - (*ptr) < 4 ) {
+    libspectrum_print_error( LIBSPECTRUM_ERROR_CORRUPT,
+			     "%s: not enough data in buffer", __func__ );
+    return LIBSPECTRUM_ERROR_CORRUPT;
+  }
+
+  length = libspectrum_read_dword( ptr );
+
+  blockend = *ptr + length;
+
+  /* Sanity check */
+  if( length < 14 ) {
+    libspectrum_print_error( LIBSPECTRUM_ERROR_CORRUPT,
+			     "%s: length less than minimum", __func__ );
+    return LIBSPECTRUM_ERROR_CORRUPT;
+  }
+
+  /* Check this much data exists */
+  if( end - (*ptr) < length ) {
+    libspectrum_print_error( LIBSPECTRUM_ERROR_CORRUPT,
+			     "%s: not enough data in buffer", __func__ );
+    return LIBSPECTRUM_ERROR_CORRUPT;
+  }
+
+  error =
+    libspectrum_tape_block_alloc( &block,
+				  LIBSPECTRUM_TAPE_BLOCK_GENERALISED_DATA );
+  if( error ) return error;
+
+  libspectrum_tape_block_zero( block );
+
+  libspectrum_tape_block_set_pause( block, (*ptr)[0] + (*ptr)[1] * 0x100 );
+  (*ptr) += 2;
+
+  error = libspectrum_tape_block_read_symbol_table_parameters( block, 1, ptr );
+  if( error ) { libspectrum_tape_block_free( block ); return error; }
+
+  error = libspectrum_tape_block_read_symbol_table_parameters( block, 0, ptr );
+  if( error ) { libspectrum_tape_block_free( block ); return error; }
+
+  length -= 14;
+
+  ptr2 = *ptr;
+
+  table = libspectrum_tape_block_pilot_table( block );
+  error = libspectrum_tape_block_read_symbol_table( table, ptr, length );
+  if( error ) { libspectrum_tape_block_free( block ); return error; }
+
+  length -= ptr2 - *ptr;
+
+  symbol_count = libspectrum_tape_generalised_data_symbol_table_symbols_in_block( table );
+
+  if( length < 3 * symbol_count ) {
+    libspectrum_tape_block_free( block );
+    libspectrum_print_error( LIBSPECTRUM_ERROR_CORRUPT,
+			     "%s: not enough data in buffer", __func__ );
+    return LIBSPECTRUM_ERROR_CORRUPT;
+  }
+
+  symbols = malloc( symbol_count * sizeof( *symbols ) );
+  if( !symbols ) {
+    libspectrum_tape_block_free( block );
+    libspectrum_print_error( LIBSPECTRUM_ERROR_MEMORY, "%s:%d", __func__,
+			     __LINE__ );
+    return LIBSPECTRUM_ERROR_MEMORY;
+  }
+
+  repeats = malloc( symbol_count * sizeof( *repeats ) );
+  if( !repeats ) {
+    libspectrum_tape_block_free( block );
+    libspectrum_print_error( LIBSPECTRUM_ERROR_MEMORY, "%s:%d", __func__,
+			     __LINE__ );
+    return LIBSPECTRUM_ERROR_MEMORY;
+  }
+
+  for( i = 0; i < symbol_count; i++ ) {
+    symbols[ i ] = **ptr; (*ptr)++;
+    repeats[ i ] = (*ptr)[0] + 0x100 * (*ptr)[1]; (*ptr) += 2;
+  }
+
+  libspectrum_tape_block_set_pilot_symbols( block, symbols );
+  libspectrum_tape_block_set_pilot_repeats( block, repeats );
+
+  length -= 3 * symbol_count;
+
+  ptr2 = *ptr;
+
+  table = libspectrum_tape_block_data_table( block );
+  libspectrum_tape_block_read_symbol_table( table, ptr, length );
+
+  length -= ptr2 - *ptr;
+
+  symbol_count2 = libspectrum_tape_generalised_data_symbol_table_symbols_in_table( table );
+
+  bits_per_symbol = ceil( log( symbol_count2 ) / M_LN2 );
+
+  libspectrum_tape_block_set_bits_per_data_symbol( block, bits_per_symbol );
+
+  symbol_count = libspectrum_tape_generalised_data_symbol_table_symbols_in_block( table );
+
+  data_count = ( ( bits_per_symbol * symbol_count ) + 7 ) / 8;
+  data_size = data_count * sizeof( *data );
+
+  data = malloc( data_size );
+  if( !data ) {
+    libspectrum_tape_block_free( block );
+    libspectrum_print_error( LIBSPECTRUM_ERROR_MEMORY, "%s:%d", __func__,
+			     __LINE__ );
+    return LIBSPECTRUM_ERROR_MEMORY;
+  }
+
+  if( *ptr + data_size > end || *ptr + data_size < *ptr ) {
+    free( data );
+    libspectrum_tape_block_free( block );
+    libspectrum_print_error( LIBSPECTRUM_ERROR_CORRUPT,
+			     "%s: data extends beyond end of block", __func__ );
+    return LIBSPECTRUM_ERROR_CORRUPT;
+  }
+
+  memcpy( data, *ptr, data_count * sizeof( *data ) );
+  *ptr += data_count;
+
+  libspectrum_tape_block_set_data( block, data );
+
+  /* Sanity check */
+  if( *ptr != blockend ) {
+    libspectrum_tape_block_free( block );
+    libspectrum_print_error( LIBSPECTRUM_ERROR_CORRUPT,
+			     "%s: sanity check failed", __func__ );
+    return LIBSPECTRUM_ERROR_CORRUPT;
+  }
+
+  error = libspectrum_tape_append_block( tape, block );
+  if( error ) { libspectrum_tape_block_free( block ); return error; }
+
+  return LIBSPECTRUM_ERROR_NONE;
+}
+
 static libspectrum_error
 tzx_read_pause( libspectrum_tape *tape, const libspectrum_byte **ptr,
 		const libspectrum_byte *end )
@@ -562,7 +735,7 @@ tzx_read_group_start( libspectrum_tape *tape, const libspectrum_byte **ptr,
 		      const libspectrum_byte *end )
 {
   libspectrum_tape_block *block;
-  libspectrum_byte *name;
+  char *name;
   libspectrum_error error;
   
   /* Check the length byte exists */
@@ -663,7 +836,7 @@ tzx_read_select( libspectrum_tape *tape, const libspectrum_byte **ptr,
 
   size_t length, count, i, j;
   int *offsets;
-  libspectrum_byte **descriptions;
+  char **descriptions;
 
   /* Check there's enough left in the buffer for the length and the count
      byte */
@@ -700,7 +873,7 @@ tzx_read_select( libspectrum_tape *tape, const libspectrum_byte **ptr,
   }
   libspectrum_tape_block_set_offsets( block, offsets );
 
-  descriptions = malloc( count * sizeof( libspectrum_byte* ) );
+  descriptions = malloc( count * sizeof( char* ) );
   if( !descriptions ) {
     free( offsets ); free( block );
     libspectrum_print_error( LIBSPECTRUM_ERROR_MEMORY,
@@ -775,7 +948,7 @@ tzx_read_comment( libspectrum_tape *tape, const libspectrum_byte **ptr,
 		  const libspectrum_byte *end )
 {
   libspectrum_tape_block* block;
-  libspectrum_byte *text;
+  char *text;
   libspectrum_error error;
 
   /* Check the length byte exists */
@@ -807,7 +980,7 @@ tzx_read_message( libspectrum_tape *tape, const libspectrum_byte **ptr,
 		  const libspectrum_byte *end )
 {
   libspectrum_tape_block* block;
-  libspectrum_byte *text;
+  char *text;
   libspectrum_error error;
 
   /* Check the time and length byte exists */
@@ -846,7 +1019,7 @@ tzx_read_archive_info( libspectrum_tape *tape, const libspectrum_byte **ptr,
 
   size_t i, count;
   int *ids;
-  libspectrum_byte **strings;
+  char **strings;
 
   /* Check there's enough left in the buffer for the length and the count
      byte */
@@ -880,7 +1053,7 @@ tzx_read_archive_info( libspectrum_tape *tape, const libspectrum_byte **ptr,
   }
   libspectrum_tape_block_set_ids( block, ids );
 
-  strings = malloc( count * sizeof( libspectrum_byte* ) );
+  strings = malloc( count * sizeof( char* ) );
   if( !strings ) {
     free( ids ); free( block );
     libspectrum_print_error( LIBSPECTRUM_ERROR_MEMORY,
@@ -894,7 +1067,7 @@ tzx_read_archive_info( libspectrum_tape *tape, const libspectrum_byte **ptr,
     /* Must be ID byte and length byte */
     if( end - (*ptr) < 2 ) {
       size_t j;
-      for( j=0; j<i; j++ ) free( strings[i] );
+      for( j=0; j<i; j++ ) free( strings[j] );
       free( strings ); free( ids ); free( block );
       libspectrum_print_error(
         LIBSPECTRUM_ERROR_CORRUPT,
@@ -910,7 +1083,7 @@ tzx_read_archive_info( libspectrum_tape *tape, const libspectrum_byte **ptr,
     error = tzx_read_string( ptr, end, &strings[i] );
     if( error ) {
       size_t j;
-      for( j = 0; j < i; j++ ) free( strings[i] );
+      for( j = 0; j < i; j++ ) free( strings[j] );
       free( strings ); free( ids ); free( block );
       return error;
     }
@@ -951,6 +1124,7 @@ tzx_read_hardware( libspectrum_tape *tape, const libspectrum_byte **ptr,
 
   /* Check there's enough data in the buffer for all the data */
   if( end - (*ptr) < 3 * (ptrdiff_t)count ) {
+    free( block );
     libspectrum_print_error( LIBSPECTRUM_ERROR_CORRUPT,
 			     "tzx_read_hardware: not enough data in buffer" );
     return LIBSPECTRUM_ERROR_CORRUPT;
@@ -1003,7 +1177,8 @@ tzx_read_custom( libspectrum_tape *tape, const libspectrum_byte **ptr,
 		 const libspectrum_byte *end )
 {
   libspectrum_tape_block* block;
-  libspectrum_byte *description, *data; size_t length;
+  char *description;
+  libspectrum_byte* data; size_t length;
   libspectrum_error error;
 
   /* Check the description (16) and length bytes (4) exist */
@@ -1025,7 +1200,7 @@ tzx_read_custom( libspectrum_tape *tape, const libspectrum_byte **ptr,
 
   /* Read in the data */
   error = tzx_read_data( ptr, end, &length, 4, &data );
-  if( error ) { free( block ); return error; }
+  if( error ) { free( description ); free( block ); return error; }
   libspectrum_tape_block_set_data_length( block, length );
   libspectrum_tape_block_set_data( block, data );
 
@@ -1041,16 +1216,18 @@ tzx_read_custom( libspectrum_tape *tape, const libspectrum_byte **ptr,
 static libspectrum_error
 tzx_read_concat( const libspectrum_byte **ptr, const libspectrum_byte *end )
 {
+  size_t signature_length = strlen( libspectrum_tzx_signature );
+
   /* Check there's enough data left; the -1 here is because we've already
      read the first byte of the signature as the block ID */
-  if( end - (*ptr) < (ptrdiff_t)strlen( signature ) + 2 - 1 ) {
+  if( end - (*ptr) < (ptrdiff_t)signature_length + 2 - 1 ) {
     libspectrum_print_error( LIBSPECTRUM_ERROR_CORRUPT,
 			     "tzx_read_concat: not enough data in buffer" );
     return LIBSPECTRUM_ERROR_CORRUPT;
   }
 
   /* Skip the data */
-  (*ptr) += strlen( signature ) + 2 - 1;
+  *ptr += signature_length + 2 - 1;
 
   return LIBSPECTRUM_ERROR_NONE;
 }
@@ -1096,17 +1273,22 @@ tzx_read_data( const libspectrum_byte **ptr, const libspectrum_byte *end,
     return LIBSPECTRUM_ERROR_CORRUPT;
   }
 
-  /* Allocate memory for the data */
-  (*data) =
-    (libspectrum_byte*)malloc( (*length+padding) * sizeof(libspectrum_byte) );
-  if( (*data) == NULL ) {
-    libspectrum_print_error( LIBSPECTRUM_ERROR_MEMORY,
-			     "tzx_read_data: out of memory" );
-    return LIBSPECTRUM_ERROR_MEMORY;
-  }
+  /* Allocate memory for the data; the check for *length is to avoid
+     the implementation-defined of malloc( 0 ) */
+  if( *length || padding ) {
 
-  /* Copy the block data across, and move along */
-  memcpy( (*data), (*ptr), (*length) ); (*ptr) += (*length);
+    *data = malloc( ( *length + padding ) * sizeof( libspectrum_byte ) );
+    if( !*data ) {
+      libspectrum_print_error( LIBSPECTRUM_ERROR_MEMORY,
+			       "tzx_read_data: out of memory" );
+      return LIBSPECTRUM_ERROR_MEMORY;
+    }
+
+    /* Copy the block data across, and move along */
+    memcpy( *data, *ptr, *length ); *ptr += *length;
+  } else {
+    *data = NULL;
+  }
 
   return LIBSPECTRUM_ERROR_NONE;
 
@@ -1114,13 +1296,13 @@ tzx_read_data( const libspectrum_byte **ptr, const libspectrum_byte *end,
 
 static libspectrum_error
 tzx_read_string( const libspectrum_byte **ptr, const libspectrum_byte *end,
-		 libspectrum_byte **dest )
+		 char **dest )
 {
   size_t length;
   libspectrum_error error;
-  libspectrum_byte *ptr2;
+  char *ptr2;
 
-  error = tzx_read_data( ptr, end, &length, -1, dest );
+  error = tzx_read_data( ptr, end, &length, -1, (libspectrum_byte**)dest );
   if( error ) return error;
   
   /* Null terminate the string */
