@@ -2,7 +2,7 @@
    Copyright (c) 2001, 2002 Philip Kendall, Darren Salt
    Copyright (c) 2003 Fredrick Meunier
 
-   $Id: warajevo_read.c,v 1.4 2003/11/27 21:45:25 pak21 Exp $
+   $Id: warajevo_read.c 3070 2007-07-20 11:06:40Z fredm $
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -14,14 +14,13 @@
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
 
-   You should have received a copy of the GNU General Public License
-   along with this program; if not, write to the Free Software
-   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
+   You should have received a copy of the GNU General Public License along
+   with this program; if not, write to the Free Software Foundation, Inc.,
+   51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 
    Author contact information:
 
-   E-mail: pak21-fuse@srcf.ucam.org
-   Postal address: 15 Crescent Road, Wokingham, Berks, RG40 2DB, England
+   E-mail: philip-fuse@shadowmagic.org.uk
 
 */
 
@@ -168,8 +167,8 @@ lsb2word( const libspectrum_byte *mem )
 /* The main load function */
 
 libspectrum_error
-libspectrum_warajevo_read( libspectrum_tape *tape,
-			   const libspectrum_byte *buffer, size_t length )
+internal_warajevo_read( libspectrum_tape *tape,
+			const libspectrum_byte *buffer, size_t length )
 {
   libspectrum_error error;
   const libspectrum_byte *ptr, *end;
@@ -206,12 +205,31 @@ libspectrum_warajevo_read( libspectrum_tape *tape,
   return LIBSPECTRUM_ERROR_NONE;
 }
 
+libspectrum_error
+libspectrum_warajevo_read( libspectrum_tape *tape,
+			   const libspectrum_byte *buffer, size_t length )
+{
+  return internal_warajevo_read( tape, buffer, length );
+}
+
 static libspectrum_error
 get_next_block( size_t *offset, const libspectrum_byte *buffer,
 		const libspectrum_byte *end, libspectrum_tape *tape )
 {
   int error;
-  libspectrum_dword next_block = lsb2dword( buffer + 4 + *offset );
+  libspectrum_dword next_block;
+
+  /* Check we have enough data, and check for pointer wrap */
+  if( buffer + 8 + *offset > end || buffer + *offset < buffer || 
+      buffer + 8 + *offset < buffer ) {
+    libspectrum_print_error(
+      LIBSPECTRUM_ERROR_CORRUPT,
+      "libspectrum_warajevo_read: not enough data in buffer"
+    );
+    return LIBSPECTRUM_ERROR_CORRUPT;
+  }
+
+  next_block = lsb2dword( buffer + 4 + *offset );
 
   /* Check for EOF marker */
   if( next_block == warajevo_signature ) {
