@@ -1,7 +1,7 @@
 /* z80.c: Routines for handling .z80 snapshots
-   Copyright (c) 2001-2004 Philip Kendall, Darren Salt
+   Copyright (c) 2001-2004 Philip Kendall, Darren Salt, Fredrick Meunier
 
-   $Id: z80.c,v 1.75 2004/07/15 20:02:20 pak21 Exp $
+   $Id: z80.c 3409 2007-12-05 02:58:49Z fredm $
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -13,14 +13,13 @@
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
 
-   You should have received a copy of the GNU General Public License
-   along with this program; if not, write to the Free Software
-   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
+   You should have received a copy of the GNU General Public License along
+   with this program; if not, write to the Free Software Foundation, Inc.,
+   51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 
    Author contact information:
 
-   E-mail: pak21-fuse@srcf.ucam.org
-   Postal address: 15 Crescent Road, Wokingham, Berks, RG40 2DB, England
+   E-mail: philip-fuse@shadowmagic.org.uk
 
 */
 
@@ -47,6 +46,12 @@ static const int LIBSPECTRUM_Z80_HEADER_LENGTH = 30;
 /* The constants used for each machine type */
 enum {
 
+  /* v1 constants */
+  Z80_JOYSTICK_CURSOR_V1 = 0,
+  Z80_JOYSTICK_KEMPSTON_V1 = 1,
+  Z80_JOYSTICK_SINCLAIR_LEFT_V1 = 2,
+  Z80_JOYSTICK_SINCLAIR_RIGHT_V1 = 3,
+
   /* v2 constants */
   Z80_MACHINE_48_V2 = 0,
   Z80_MACHINE_48_IF1_V2 = 1,
@@ -63,6 +68,15 @@ enum {
   Z80_MACHINE_128_IF1 = 5,
   Z80_MACHINE_128_MGT = 6,
 
+  Z80_MGT_DISCIPLE_EPSON = 0,
+  Z80_MGT_DISCIPLE_HP = 1,
+  Z80_MGT_PLUSD = 16,
+
+  Z80_JOYSTICK_CURSOR_V3 = 0,
+  Z80_JOYSTICK_KEMPSTON_V3 = 1,
+  Z80_JOYSTICK_USER_DEFINED_V3 = 2,
+  Z80_JOYSTICK_SINCLAIR_RIGHT_V3 = 3,
+
   /* Extensions */
   Z80_MACHINE_PLUS3 = 7,
   Z80_MACHINE_PLUS3_XZX_ERROR = 8,
@@ -79,6 +93,19 @@ enum {
   Z80_MACHINE_FIRST_EXTENSION = Z80_MACHINE_PLUS3,
 };
 
+/* Constants representing a Sinclair Interface II port 1 joystick as a .z80
+   user defined joystick */
+static const libspectrum_word if2_left_map_l = 0x0f03;
+static const libspectrum_word if2_left_map_r = 0x0803;
+static const libspectrum_word if2_left_map_d = 0x0403;
+static const libspectrum_word if2_left_map_u = 0x0203;
+static const libspectrum_word if2_left_map_f = 0x0103;
+static const libspectrum_word if2_left_l     = '1';
+static const libspectrum_word if2_left_r     = '2';
+static const libspectrum_word if2_left_d     = '3';
+static const libspectrum_word if2_left_u     = '4';
+static const libspectrum_word if2_left_f     = '5';
+
 /* The signature used to designate the .slt extensions */
 static libspectrum_byte slt_signature[] = "\0\0\0SLT";
 static size_t slt_signature_length = 6;
@@ -87,14 +114,26 @@ static libspectrum_error
 read_header( const libspectrum_byte *buffer, libspectrum_snap *snap,
 	     const libspectrum_byte **data, int *version, int *compressed );
 static libspectrum_error
-get_machine_type( libspectrum_snap *snap, libspectrum_byte type, int version );
+get_machine_type( libspectrum_snap *snap, libspectrum_byte type,
+                  libspectrum_byte mgt_type, int version );
 static libspectrum_error
 get_machine_type_v2( libspectrum_snap *snap, libspectrum_byte type );
 static libspectrum_error
-get_machine_type_v3( libspectrum_snap *snap, libspectrum_byte type );
+get_machine_type_v3( libspectrum_snap *snap, libspectrum_byte type,
+                     libspectrum_byte mgt_type );
 static libspectrum_error
 get_machine_type_extension( libspectrum_snap *snap, libspectrum_byte type );
 static libspectrum_error
+get_joystick_type( libspectrum_snap *snap, 
+                   const libspectrum_byte *custom_joystick_base,
+                   libspectrum_byte type, int version );
+static libspectrum_error
+get_joystick_type_v1( libspectrum_snap *snap, libspectrum_byte type );
+static libspectrum_error
+get_joystick_type_v3( libspectrum_snap *snap,
+                      const libspectrum_byte *custom_joystick_base,
+                      libspectrum_byte type );
+static libspectrum_error
 read_blocks( const libspectrum_byte *buffer, size_t buffer_length,
 	     libspectrum_snap *snap, int version, int compressed );
 static libspectrum_error
@@ -119,7 +158,7 @@ write_header( libspectrum_byte **buffer, libspectrum_byte **ptr,
 	      size_t *length, int *flags, libspectrum_snap *snap );
 static libspectrum_error
 write_base_header( libspectrum_byte **buffer, libspectrum_byte **ptr,
-		   size_t *length, libspectrum_snap *snap );
+		   size_t *length, int *flags, libspectrum_snap *snap );
 static libspectrum_error
 write_extended_header( libspectrum_byte **buffer, libspectrum_byte **ptr,
 		       size_t *length, int *flags, libspectrum_snap *snap );
@@ -160,10 +199,17 @@ enum slt_type {
 libspectrum_error
 libspectrum_z80_read( libspectrum_snap *snap,
 	              const libspectrum_byte *buffer, size_t buffer_length )
+{
+  return internal_z80_read( snap, buffer, buffer_length );
+}
+
+libspectrum_error
+internal_z80_read( libspectrum_snap *snap,
+		   const libspectrum_byte *buffer, size_t buffer_length )
 {
   libspectrum_error error;
   const libspectrum_byte *data;
-  int version, compressed;
+  int version, compressed = 1;
 
   error = read_header( buffer, snap, &data, &version, &compressed );
   if( error != LIBSPECTRUM_ERROR_NONE ) return error;
@@ -172,6 +218,11 @@ libspectrum_z80_read( libspectrum_snap *snap,
 		       version, compressed );
   if( error != LIBSPECTRUM_ERROR_NONE ) return error;
 
+  libspectrum_snap_set_beta_paged( snap, 0 );
+
+  if( libspectrum_snap_interface1_active( snap ) )
+    libspectrum_snap_set_interface1_drive_count( snap, 8 );
+
   return LIBSPECTRUM_ERROR_NONE;
 }
 
@@ -205,6 +256,7 @@ read_header( const libspectrum_byte *buffer, libspectrum_snap *snap,
   libspectrum_snap_set_iff1( snap, header[27] ? 1 : 0 );
   libspectrum_snap_set_iff2( snap, header[28] ? 1 : 0 );
   libspectrum_snap_set_im( snap, header[29] & 0x03 );
+  libspectrum_snap_set_issue2( snap, header[29] & 0x04 );
 
   libspectrum_snap_set_out_ula( snap, (header[12]&0x0e) >> 1 );
 
@@ -237,11 +289,22 @@ read_header( const libspectrum_byte *buffer, libspectrum_snap *snap,
 
     extra_header = buffer + LIBSPECTRUM_Z80_HEADER_LENGTH + 2;
 
+    error = get_joystick_type( snap,
+                               extra_header + 32,
+                               ( header[29] & 0xc0 ) >> 6,
+                               *version
+                             );
+    if( error ) return error;
+
     libspectrum_snap_set_pc( snap, extra_header[0] + extra_header[1] * 0x100 );
 
-    error = get_machine_type( snap, extra_header[2], *version );
+    error = get_machine_type( snap, extra_header[2], extra_header[51], *version );
     if( error ) return error;
 
+    if( extra_header[27] && libspectrum_snap_plusd_active( snap ) ) {
+      libspectrum_snap_set_plusd_paged( snap, 1 );
+    }
+
     if( *version >= 3 ) {
 
       quarter_tstates =
@@ -331,6 +394,10 @@ read_header( const libspectrum_byte *buffer, libspectrum_snap *snap,
     libspectrum_snap_set_machine( snap, LIBSPECTRUM_MACHINE_48 );
     *version = 1;
 
+    libspectrum_snap_set_joystick_active_count( snap, 1 );
+    error = get_joystick_type_v1( snap, ( header[29] & 0xc0 ) >> 6 );
+    if( error ) return error;
+
     /* Need to flag this for later */
     *compressed = ( header[12] & 0x20 ) ? 1 : 0;
 
@@ -339,11 +406,11 @@ read_header( const libspectrum_byte *buffer, libspectrum_snap *snap,
   }
 
   return LIBSPECTRUM_ERROR_NONE;
-
 }
 
 static libspectrum_error
-get_machine_type( libspectrum_snap *snap, libspectrum_byte type, int version )
+get_machine_type( libspectrum_snap *snap, libspectrum_byte type,
+                  libspectrum_byte mgt_type, int version )
 {
   libspectrum_error error;
 
@@ -356,7 +423,7 @@ get_machine_type( libspectrum_snap *snap, libspectrum_byte type, int version )
       break;
 
     case 3:
-      error = get_machine_type_v3( snap, type ); if( error ) return error;
+      error = get_machine_type_v3( snap, type, mgt_type ); if( error ) return error;
       break;
 
     default:
@@ -379,11 +446,15 @@ get_machine_type_v2( libspectrum_snap *snap, libspectrum_byte type )
   switch( type ) {
 
   case Z80_MACHINE_48_V2:
-  case Z80_MACHINE_48_IF1_V2:
   case Z80_MACHINE_48_SAMRAM_V2:
     libspectrum_snap_set_machine( snap, LIBSPECTRUM_MACHINE_48 ); break;
+  case Z80_MACHINE_48_IF1_V2:
+    libspectrum_snap_set_interface1_active( snap, 1 );
+    libspectrum_snap_set_machine( snap, LIBSPECTRUM_MACHINE_48 ); break;
   case Z80_MACHINE_128_V2:
+    libspectrum_snap_set_machine( snap, LIBSPECTRUM_MACHINE_128 ); break;
   case Z80_MACHINE_128_IF1_V2:
+    libspectrum_snap_set_interface1_active( snap, 1 );
     libspectrum_snap_set_machine( snap, LIBSPECTRUM_MACHINE_128 ); break;
   default:
     libspectrum_print_error( LIBSPECTRUM_ERROR_UNKNOWN,
@@ -396,19 +467,57 @@ get_machine_type_v2( libspectrum_snap *snap, libspectrum_byte type )
 }
 
 static libspectrum_error
-get_machine_type_v3( libspectrum_snap *snap, libspectrum_byte type )
+get_mgt_type( libspectrum_snap *snap, libspectrum_byte type )
 {
   switch( type ) {
     
+  case Z80_MGT_DISCIPLE_EPSON:
+  case Z80_MGT_DISCIPLE_HP:
+    break;
+
+  case Z80_MGT_PLUSD:
+    libspectrum_snap_set_plusd_active( snap, 1 );
+    break;
+
+  default:
+    libspectrum_print_error( LIBSPECTRUM_ERROR_UNKNOWN,
+			     "%s:get_mgt_type: unknown mgt type %d",
+			     __FILE__, type );
+    return LIBSPECTRUM_ERROR_UNKNOWN;
+  }
+
+  return LIBSPECTRUM_ERROR_NONE;
+}
+
+static libspectrum_error
+get_machine_type_v3( libspectrum_snap *snap, libspectrum_byte type,
+                     libspectrum_byte mgt_type )
+{
+  libspectrum_error error;
+
+  switch( type ) {
+    
   case Z80_MACHINE_48:
-  case Z80_MACHINE_48_IF1:
   case Z80_MACHINE_48_SAMRAM:
-  case Z80_MACHINE_48_MGT:
     libspectrum_snap_set_machine( snap, LIBSPECTRUM_MACHINE_48 ); break;
+  case Z80_MACHINE_48_IF1:
+    libspectrum_snap_set_interface1_active( snap, 1 );
+    libspectrum_snap_set_machine( snap, LIBSPECTRUM_MACHINE_48 ); break;
+  case Z80_MACHINE_48_MGT:
+    libspectrum_snap_set_machine( snap, LIBSPECTRUM_MACHINE_48 );
+    error = get_mgt_type( snap, mgt_type );
+    if( error != LIBSPECTRUM_ERROR_NONE ) return error;
+    break;
   case Z80_MACHINE_128:
-  case Z80_MACHINE_128_IF1:
-  case Z80_MACHINE_128_MGT:
     libspectrum_snap_set_machine( snap, LIBSPECTRUM_MACHINE_128 ); break;
+  case Z80_MACHINE_128_IF1:
+    libspectrum_snap_set_interface1_active( snap, 1 );
+    libspectrum_snap_set_machine( snap, LIBSPECTRUM_MACHINE_128 ); break;
+  case Z80_MACHINE_128_MGT:
+    libspectrum_snap_set_machine( snap, LIBSPECTRUM_MACHINE_128 );
+    error = get_mgt_type( snap, mgt_type );
+    if( error != LIBSPECTRUM_ERROR_NONE ) return error;
+    break;
   default:
     libspectrum_print_error( LIBSPECTRUM_ERROR_UNKNOWN,
 			     "%s:get_machine_type: unknown v3 machine type %d",
@@ -438,8 +547,9 @@ get_machine_type_extension( libspectrum_snap *snap, libspectrum_byte type )
   case Z80_MACHINE_TC2048:
     libspectrum_snap_set_machine( snap, LIBSPECTRUM_MACHINE_TC2048 ); break;
   case Z80_MACHINE_TC2068:
-  case Z80_MACHINE_TS2068: /* Load TS2068 snaps as TC2068 for now */
     libspectrum_snap_set_machine( snap, LIBSPECTRUM_MACHINE_TC2068 ); break;
+  case Z80_MACHINE_TS2068:
+    libspectrum_snap_set_machine( snap, LIBSPECTRUM_MACHINE_TS2068 ); break;
   default:
     libspectrum_print_error(
       LIBSPECTRUM_ERROR_UNKNOWN,
@@ -451,6 +561,115 @@ get_machine_type_extension( libspectrum_snap *snap, libspectrum_byte type )
   return LIBSPECTRUM_ERROR_NONE;
 }
 
+static libspectrum_error
+get_joystick_type( libspectrum_snap *snap, 
+                   const libspectrum_byte *custom_joystick_base,
+                   libspectrum_byte type, int version )
+{
+  libspectrum_error error;
+
+  libspectrum_snap_set_joystick_active_count( snap, 1 );
+  libspectrum_snap_set_joystick_inputs( snap, 0,
+                                        LIBSPECTRUM_JOYSTICK_INPUT_KEYBOARD |
+                                        LIBSPECTRUM_JOYSTICK_INPUT_JOYSTICK_1 );
+
+  switch( version ) {
+
+  case 1:
+  case 2:
+    error = get_joystick_type_v1( snap, type ); if( error ) return error;
+    break;
+  case 3:
+    error = get_joystick_type_v3( snap, custom_joystick_base, type );
+    if( error ) return error;
+    break;
+
+  default:
+    libspectrum_print_error( LIBSPECTRUM_ERROR_LOGIC,
+                             "%s:get_joystick_type: unknown version %d",
+                             __FILE__, version );
+    return LIBSPECTRUM_ERROR_LOGIC;
+
+  }
+
+  return LIBSPECTRUM_ERROR_NONE;
+}
+
+static libspectrum_error
+get_joystick_type_v1( libspectrum_snap *snap, libspectrum_byte type )
+{
+  switch( type ) {
+
+  case Z80_JOYSTICK_CURSOR_V1:
+    libspectrum_snap_set_joystick_list( snap, 0, LIBSPECTRUM_JOYSTICK_CURSOR );
+    break;
+  case Z80_JOYSTICK_KEMPSTON_V1:
+    libspectrum_snap_set_joystick_list( snap, 0, LIBSPECTRUM_JOYSTICK_KEMPSTON );
+    break;
+  case Z80_JOYSTICK_SINCLAIR_LEFT_V1:
+    libspectrum_snap_set_joystick_list( snap, 0, LIBSPECTRUM_JOYSTICK_SINCLAIR_2 );
+    break;
+  case Z80_JOYSTICK_SINCLAIR_RIGHT_V1:
+    libspectrum_snap_set_joystick_list( snap, 0, LIBSPECTRUM_JOYSTICK_SINCLAIR_1 );
+    break;
+
+  default:
+    libspectrum_print_error( LIBSPECTRUM_ERROR_UNKNOWN,
+			     "%s:get_joystick_type_v1: unknown v1 "
+                             "joystick type %d",
+			     __FILE__, type );
+    return LIBSPECTRUM_ERROR_UNKNOWN;
+  }
+
+  return LIBSPECTRUM_ERROR_NONE;
+}
+
+static libspectrum_error
+get_joystick_type_v3( libspectrum_snap *snap,
+                      const libspectrum_byte *custom_keys_base,
+                      libspectrum_byte type )
+{
+  switch( type ) {
+
+  case Z80_JOYSTICK_CURSOR_V3:
+    libspectrum_snap_set_joystick_list( snap, 0, LIBSPECTRUM_JOYSTICK_CURSOR );
+    break;
+  case Z80_JOYSTICK_KEMPSTON_V3:
+    libspectrum_snap_set_joystick_list( snap, 0, LIBSPECTRUM_JOYSTICK_KEMPSTON );
+    break;
+  case Z80_JOYSTICK_USER_DEFINED_V3:
+    /* User defined mapping, check to see if keys match Sinclair Interface 2
+       left port and return that if possible */
+    /* First 10 bytes are spectrum keyboard mappings for user defined keys
+       in this order: left, right, down, up, fire */
+    if( custom_keys_base[0] + custom_keys_base[1] * 0x100 == if2_left_map_l &&
+        custom_keys_base[2] + custom_keys_base[3] * 0x100 == if2_left_map_r &&
+        custom_keys_base[4] + custom_keys_base[5] * 0x100 == if2_left_map_d &&
+        custom_keys_base[6] + custom_keys_base[7] * 0x100 == if2_left_map_u &&
+        custom_keys_base[8] + custom_keys_base[9] * 0x100 == if2_left_map_f ) {
+      libspectrum_snap_set_joystick_list( snap, 0,
+                                          LIBSPECTRUM_JOYSTICK_SINCLAIR_2 );
+    } else {
+      libspectrum_snap_set_joystick_active_count( snap, 0 );
+      libspectrum_snap_set_joystick_inputs( snap, 0, 0 );
+    }
+    break;
+  case Z80_JOYSTICK_SINCLAIR_RIGHT_V3:
+    libspectrum_snap_set_joystick_list( snap, 0,
+                                        LIBSPECTRUM_JOYSTICK_SINCLAIR_1 );
+    break;
+
+  default:
+    libspectrum_print_error( LIBSPECTRUM_ERROR_UNKNOWN,
+			     "%s:get_joystick_type_v3: unknown v3 joystick "
+                             "type %d",
+			     __FILE__, type );
+    return LIBSPECTRUM_ERROR_UNKNOWN;
+  }
+
+  return LIBSPECTRUM_ERROR_NONE;
+}
+
 static libspectrum_error
 read_blocks( const libspectrum_byte *buffer, size_t buffer_length,
 	     libspectrum_snap *snap, int version, int compressed )
@@ -676,9 +895,34 @@ read_block( const libspectrum_byte *buffer, libspectrum_snap *snap,
     if( page <= 0 || page > 18 ) {
       libspectrum_print_error( LIBSPECTRUM_ERROR_UNKNOWN,
 			       "read_block: unknown page %d", page );
+      free( uncompressed );
       return LIBSPECTRUM_ERROR_UNKNOWN;
     }
 
+    /* If it is an Interface 1 ROM page put it in the appropriate structure */
+    if( page == 1 && libspectrum_snap_interface1_active( snap ) ) {
+      libspectrum_byte *chunk = malloc( 0x4000 );
+      memcpy( chunk, uncompressed, 0x4000 );
+      libspectrum_snap_set_interface1_custom_rom( snap, 1 );
+      libspectrum_snap_set_interface1_rom( snap, 0, chunk );
+      libspectrum_snap_set_interface1_rom_length( snap, 0, 0x4000 );
+      free( uncompressed );
+      return LIBSPECTRUM_ERROR_NONE;
+    }
+
+    /* If it is a +D ROM/RAM page put it in the appropriate structures */
+    if( page == 1 && libspectrum_snap_plusd_active( snap ) ) {
+      /* Bottom 8K of page is +D ROM, upper 8K is +D RAM */
+      libspectrum_byte *chunk = malloc( 0x2000 );
+      memcpy( chunk, uncompressed, 0x2000 );
+      libspectrum_snap_set_plusd_rom( snap, 0, chunk );
+      chunk = malloc( 0x2000 );
+      memcpy( chunk, uncompressed + 0x2000, 0x2000 );
+      libspectrum_snap_set_plusd_ram( snap, 0, chunk );
+      free( uncompressed );
+      return LIBSPECTRUM_ERROR_NONE;
+    }
+
     /* If it's a ROM page, just throw it away */
     if( page < 3 ) {
       free( uncompressed );
@@ -918,6 +1162,10 @@ libspectrum_z80_write2( libspectrum_byte **buffer, size_t *length,
   if( libspectrum_snap_halted( snap ) )
     *out_flags |= LIBSPECTRUM_FLAG_SNAPSHOT_MINOR_INFO_LOSS;
 
+  /* .z80 format doesn't store +D info well */
+  if( libspectrum_snap_plusd_active( snap ) )
+    *out_flags |= LIBSPECTRUM_FLAG_SNAPSHOT_MINOR_INFO_LOSS;
+
   /* .z80 format doesn't store the ZXATASP or ZXCF info */
   if( libspectrum_snap_zxatasp_active( snap ) )
     *out_flags |= LIBSPECTRUM_FLAG_SNAPSHOT_MAJOR_INFO_LOSS;
@@ -938,8 +1186,8 @@ libspectrum_z80_write2( libspectrum_byte **buffer, size_t *length,
 
   error = write_pages(
     buffer, &ptr, length, snap,
-    !( in_flags & LIBSPECTRUM_FLAG_SNAPSHOT_NO_COMPRESSION )
-  );
+    !( in_flags & LIBSPECTRUM_FLAG_SNAPSHOT_NO_COMPRESSION) +
+    ( in_flags & LIBSPECTRUM_FLAG_SNAPSHOT_ALWAYS_COMPRESS ));
   if( error != LIBSPECTRUM_ERROR_NONE ) return error;
 
   (*length) = ptr - *buffer;
@@ -953,7 +1201,7 @@ write_header( libspectrum_byte **buffer, libspectrum_byte **ptr,
 {
   libspectrum_error error;
 
-  error = write_base_header( buffer, ptr, length, snap );
+  error = write_base_header( buffer, ptr, length, flags, snap );
   if( error != LIBSPECTRUM_ERROR_NONE ) return error;
 
   error = write_extended_header( buffer, ptr, length, flags, snap );
@@ -964,9 +1212,10 @@ write_header( libspectrum_byte **buffer, libspectrum_byte **ptr,
 
 static libspectrum_error
 write_base_header( libspectrum_byte **buffer, libspectrum_byte **ptr,
-		   size_t *length, libspectrum_snap *snap )
+		   size_t *length, int *flags, libspectrum_snap *snap )
 {
   libspectrum_error error;
+  libspectrum_byte joystick_flags = 0;
 
   error = libspectrum_make_room( buffer, LIBSPECTRUM_Z80_HEADER_LENGTH, ptr,
 				 length );
@@ -997,7 +1246,39 @@ write_base_header( libspectrum_byte **buffer, libspectrum_byte **ptr,
 
   *(*ptr)++ = libspectrum_snap_iff1( snap ) ? 0xff : 0x00;
   *(*ptr)++ = libspectrum_snap_iff2( snap ) ? 0xff : 0x00;
-  *(*ptr)++ = libspectrum_snap_im( snap );
+
+  /* Exactly one active joystick can be represented in a .z80 file */
+  if( libspectrum_snap_joystick_active_count( snap ) != 1 )
+    *flags |= LIBSPECTRUM_FLAG_SNAPSHOT_MINOR_INFO_LOSS;
+
+  switch( libspectrum_snap_joystick_list( snap, 0 ) ) {
+  case LIBSPECTRUM_JOYSTICK_CURSOR:
+    joystick_flags = Z80_JOYSTICK_CURSOR_V3;
+    break;
+  case LIBSPECTRUM_JOYSTICK_KEMPSTON:
+    joystick_flags = Z80_JOYSTICK_KEMPSTON_V3;
+    break;
+  case LIBSPECTRUM_JOYSTICK_SINCLAIR_1:
+    joystick_flags = Z80_JOYSTICK_SINCLAIR_RIGHT_V3;
+    break;
+  case LIBSPECTRUM_JOYSTICK_SINCLAIR_2:
+    /* Write Z80_JOYSTICK_USER_DEFINED_V3 and custom key mapping corresponding
+       to Sinclair Interface II right port */
+    joystick_flags = Z80_JOYSTICK_USER_DEFINED_V3;
+    break;
+
+  default:
+    /* LIBSPECTRUM_JOYSTICK_NONE,
+       LIBSPECTRUM_JOYSTICK_TIMEX_1,
+       LIBSPECTRUM_JOYSTICK_TIMEX_2,
+       LIBSPECTRUM_JOYSTICK_FULLER, */
+    *flags |= LIBSPECTRUM_FLAG_SNAPSHOT_MINOR_INFO_LOSS;
+  }
+  joystick_flags <<= 6;
+
+  *(*ptr)++ = ( libspectrum_snap_im( snap ) & 0x03 ) +
+              ( libspectrum_snap_issue2( snap ) ? 0x04 : 0x00 ) +
+              joystick_flags;
 
   return LIBSPECTRUM_ERROR_NONE;
 }
@@ -1025,29 +1306,53 @@ write_extended_header( libspectrum_byte **buffer, libspectrum_byte **ptr,
   switch( libspectrum_snap_machine( snap ) ) {
   case LIBSPECTRUM_MACHINE_16:
   case LIBSPECTRUM_MACHINE_48:
-    *(*ptr)++ = Z80_MACHINE_48; break;
+    if( libspectrum_snap_plusd_active( snap ) ) {
+      *(*ptr)++ = Z80_MACHINE_48_MGT;
+    } else if( libspectrum_snap_interface1_active( snap ) ) {
+      *(*ptr)++ = Z80_MACHINE_48_IF1;
+    } else {
+      *(*ptr)++ = Z80_MACHINE_48;
+    }
+    break;
   case LIBSPECTRUM_MACHINE_SE:
     *flags |= LIBSPECTRUM_FLAG_SNAPSHOT_MAJOR_INFO_LOSS;
     /* fall through */
   case LIBSPECTRUM_MACHINE_128:
-    *(*ptr)++ = Z80_MACHINE_128; break;
+    if( libspectrum_snap_plusd_active( snap ) ) {
+      *(*ptr)++ = Z80_MACHINE_128_MGT;
+    } else if( libspectrum_snap_interface1_active( snap ) ) {
+      *(*ptr)++ = Z80_MACHINE_128_IF1;
+    } else {
+      *(*ptr)++ = Z80_MACHINE_128;
+    }
+    break;
   case LIBSPECTRUM_MACHINE_PLUS3E:
     *flags |= LIBSPECTRUM_FLAG_SNAPSHOT_MINOR_INFO_LOSS;
     /* fall through */
   case LIBSPECTRUM_MACHINE_PLUS3:
     *(*ptr)++ = Z80_MACHINE_PLUS3; break;
   case LIBSPECTRUM_MACHINE_PENT:
+  case LIBSPECTRUM_MACHINE_PENT512:
+  case LIBSPECTRUM_MACHINE_PENT1024:
     *(*ptr)++ = Z80_MACHINE_PENTAGON; break;
   case LIBSPECTRUM_MACHINE_SCORP:
     *(*ptr)++ = Z80_MACHINE_SCORPION; break;
   case LIBSPECTRUM_MACHINE_PLUS2:
+    if( libspectrum_snap_interface1_active( snap ) ) {
+      *flags |= LIBSPECTRUM_FLAG_SNAPSHOT_MINOR_INFO_LOSS;
+    }
     *(*ptr)++ = Z80_MACHINE_PLUS2; break;
   case LIBSPECTRUM_MACHINE_PLUS2A:
     *(*ptr)++ = Z80_MACHINE_PLUS2A; break;
   case LIBSPECTRUM_MACHINE_TC2048:
+    if( libspectrum_snap_interface1_active( snap ) ) {
+      *flags |= LIBSPECTRUM_FLAG_SNAPSHOT_MINOR_INFO_LOSS;
+    }
     *(*ptr)++ = Z80_MACHINE_TC2048; break;
   case LIBSPECTRUM_MACHINE_TC2068:
     *(*ptr)++ = Z80_MACHINE_TC2068; break;
+  case LIBSPECTRUM_MACHINE_TS2068:
+    *(*ptr)++ = Z80_MACHINE_TS2068; break;
 
   case LIBSPECTRUM_MACHINE_UNKNOWN:
     libspectrum_print_error( LIBSPECTRUM_ERROR_UNKNOWN,
@@ -1066,8 +1371,14 @@ write_extended_header( libspectrum_byte **buffer, libspectrum_byte **ptr,
 
   if( capabilities & LIBSPECTRUM_MACHINE_CAPABILITY_TIMEX_VIDEO ) {
     *(*ptr)++ = libspectrum_snap_out_scld_dec( snap );
+    if( libspectrum_snap_interface1_active( snap ) &&
+        libspectrum_snap_interface1_paged( snap ) )
+      *flags |= LIBSPECTRUM_FLAG_SNAPSHOT_MAJOR_INFO_LOSS;
+  } else if( libspectrum_snap_interface1_active( snap ) &&
+             libspectrum_snap_interface1_paged( snap ) ) {
+    *(*ptr)++ = 0xff;
   } else {
-    *(*ptr)++ = '\0';		/* IF1 disabled */
+    *(*ptr)++ = '\0';
   }
 
   /* Support 16K snapshots via Spectaculator's extension; see the
@@ -1095,8 +1406,18 @@ write_extended_header( libspectrum_byte **buffer, libspectrum_byte **ptr,
   *(*ptr)++ =
     ( ( libspectrum_snap_tstates( snap ) / quarter_states ) + 3 ) % 4;
 
-  /* Spectator, MGT and Multiface disabled */
-  *(*ptr)++ = '\0'; *(*ptr)++ = '\0'; *(*ptr)++ = '\0';
+  /* Spectator disabled */
+  *(*ptr)++ = '\0';
+
+  /* MGT only for +D for now */
+  if( libspectrum_snap_plusd_active( snap ) ) {
+    *(*ptr)++ = libspectrum_snap_plusd_paged( snap );
+  } else {
+    *(*ptr)++ = '\0';
+  }
+
+  /* Multiface disabled */
+  *(*ptr)++ = '\0';
 
   /* Is 0x0000 to 0x3fff RAM? True if we're in a +3 64Kb RAM
      configuration, or a Scorpion configuration with page 0 mapped in */
@@ -1119,8 +1440,37 @@ write_extended_header( libspectrum_byte **buffer, libspectrum_byte **ptr,
     *(*ptr)++ = 0x00; *(*ptr)++ = 0x00;
   }
 
-  /* Joystick settings, etc */
-  for( i=32; i<=LIBSPECTRUM_Z80_V3_LENGTH; i++ ) *(*ptr)++ = '\0';
+  /* Joystick settings */
+  if( libspectrum_snap_joystick_list( snap, 1 ) ==
+        LIBSPECTRUM_JOYSTICK_SINCLAIR_2 ) {
+    /* First 10 bytes are spectrum keyboard mappings for user defined keys
+       in this order: left, right, down, up, fire */
+    libspectrum_write_word( ptr, if2_left_map_l );
+    libspectrum_write_word( ptr, if2_left_map_r );
+    libspectrum_write_word( ptr, if2_left_map_d );
+    libspectrum_write_word( ptr, if2_left_map_u );
+    libspectrum_write_word( ptr, if2_left_map_f );
+    /* Second 10 bytes are ascii words corresponding to keys */
+    libspectrum_write_word( ptr, if2_left_l );
+    libspectrum_write_word( ptr, if2_left_r );
+    libspectrum_write_word( ptr, if2_left_d );
+    libspectrum_write_word( ptr, if2_left_u );
+    libspectrum_write_word( ptr, if2_left_f );
+  } else {
+    for( i=32; i<52; i++ ) *(*ptr)++ = '\0';
+  }
+
+  /* MGT type */
+  if( libspectrum_snap_plusd_active( snap ) ) {
+    *(*ptr)++ = Z80_MGT_PLUSD;
+    *(*ptr)++ = 0xff; /*    0-8191 ROM */
+    *(*ptr)++ = 0x00; /* 8192-9383 RAM */
+  } else {
+    for( i=52; i<55; i++ ) *(*ptr)++ = '\0';
+  }
+
+  /* etc */
+  for( i=55; i<=LIBSPECTRUM_Z80_V3_LENGTH; i++ ) *(*ptr)++ = '\0';
 
   return LIBSPECTRUM_ERROR_NONE;
 }
@@ -1135,6 +1485,37 @@ write_pages( libspectrum_byte **buffer, libspectrum_byte **ptr, size_t *length,
   int capabilities =
     libspectrum_machine_capabilities( libspectrum_snap_machine( snap ) );
 
+  /* If Interface 1 is enabled, write the Interface 1 ROM in Z80 page 1 */
+  if( libspectrum_snap_interface1_active( snap ) &&
+      libspectrum_snap_interface1_custom_rom( snap ) ) {
+    libspectrum_byte *uncompressed = calloc( 0x4000, 1 );
+    
+    memcpy( uncompressed, libspectrum_snap_interface1_rom( snap, 0 ),
+            libspectrum_snap_interface1_rom_length( snap, 0 ) );
+
+    error = write_page( buffer, ptr, length, 1, uncompressed, compress );
+    if( error != LIBSPECTRUM_ERROR_NONE ) return error;
+
+    free( uncompressed );
+  }
+
+  /* If +D is enabled, write the +D ROM and RAM in Z80 page 1 as long as we
+     have a ROM to write, presumably there is not much point saving the +D
+     RAM if we also overwrite the ROM with 0 */
+  if( libspectrum_snap_plusd_active( snap ) &&
+      libspectrum_snap_plusd_custom_rom( snap ) ) {
+    libspectrum_byte *uncompressed = malloc( 0x4000 );
+    
+    memcpy( uncompressed, libspectrum_snap_plusd_rom( snap, 0 ), 0x2000 );
+    memcpy( uncompressed + 0x2000,
+            libspectrum_snap_plusd_ram( snap, 0 ), 0x2000 );
+
+    error = write_page( buffer, ptr, length, 1, uncompressed, compress );
+    if( error != LIBSPECTRUM_ERROR_NONE ) return error;
+
+    free( uncompressed );
+  }
+
   if( !( capabilities & LIBSPECTRUM_MACHINE_CAPABILITY_128_MEMORY ) ) {
 
     error = write_page( buffer, ptr, length, 4,
@@ -1204,7 +1585,8 @@ write_page( libspectrum_byte **buffer, libspectrum_byte **ptr, size_t *length,
 
   }
 
-  if( !compress || compressed_length >= 0x4000 ) {
+  if( !compress || ( !(compress & LIBSPECTRUM_FLAG_SNAPSHOT_ALWAYS_COMPRESS) &&
+                      compressed_length >= 0x4000 ) ) {
 
     error = libspectrum_make_room( buffer, 3 + 0x4000, ptr, length );
     if( error != LIBSPECTRUM_ERROR_NONE ) return error;
