@@ -1,7 +1,7 @@
 /* zxs.c: Routines for .zxs snapshots
    Copyright (c) 1998,2003 Philip Kendall
 
-   $Id: zxs.c,v 1.4 2003/11/27 21:45:25 pak21 Exp $
+   $Id: zxs.c 2890 2007-05-26 19:31:43Z zubzero $
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -13,14 +13,13 @@
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
 
-   You should have received a copy of the GNU General Public License
-   along with this program; if not, write to the Free Software
-   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
+   You should have received a copy of the GNU General Public License along
+   with this program; if not, write to the Free Software Foundation, Inc.,
+   51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 
    Author contact information:
 
-   E-mail: pak21-fuse@srcf.ucam.org
-   Postal address: 15 Crescent Road, Wokingham, Berks, RG40 2DB, England
+   E-mail: philip-fuse@shadowmagic.org.uk
 
 */
 
@@ -28,7 +27,9 @@
 
 #include <string.h>
 
+#ifdef HAVE_ZLIB_H
 #include <zlib.h>
+#endif				/* #ifdef HAVE_ZLIB_H */
 
 #include "internals.h"
 
@@ -47,6 +48,9 @@ static libspectrum_error
 inflate_block( libspectrum_byte **uncompressed, size_t *uncompressed_length,
 	       const libspectrum_byte **compressed, size_t compressed_length )
 {
+
+#ifdef HAVE_ZLIB_H
+
   libspectrum_dword header_length, expected_crc32, actual_crc32;
   libspectrum_byte *zlib_buffer;
   unsigned long actual_length;
@@ -131,6 +135,14 @@ inflate_block( libspectrum_byte **uncompressed, size_t *uncompressed_length,
   }
 
   return LIBSPECTRUM_ERROR_NONE;
+
+#else				/* #ifdef HAVE_ZLIB_H */
+
+  /* No zlib, so can't inflate the block */
+  return LIBSPECTRUM_ERROR_UNKNOWN;
+
+#endif				/* #ifdef HAVE_ZLIB_H */
+
 }
 
 static libspectrum_error
